# encoding: utf-8
from Aspidites._vendor.contracts import contract
from Aspidites._vendor.contracts.main import parse_contract_string
from Aspidites._vendor.contracts.library import Extension

name = "helló wörld from one"

import unittest


class TestParsingNumbers(unittest.TestCase):
    def test_unicode_literal(self):
        r = parse_contract_string(u"int")
        assert r == Extension("int")

    def test_unicode_literal2(self):
        @contract(x="string")
        def f(x):
            pass

        f("")



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY2
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval
import pytest


def test_Num():
    """It allows to use number literals."""
    assert restricted_eval("42") == 42


def test_Bytes():
    """It allows to use bytes literals."""
    assert restricted_eval('b"code"') == b"code"


def test_Set():
    """It allows to use set literals."""
    assert restricted_eval("{1, 2, 3}") == set([1, 2, 3])


@pytest.mark.skipif(IS_PY2, reason="... is new in Python 3")
def test_Ellipsis():
    """It prevents using the `ellipsis` statement."""
    result = compile_restricted_exec("...")
    assert result.errors == ("Line 1: Ellipsis statements are not allowed.",)



import hypothesis
import pytest
from hypothesis import given, assume
import hypothesis.strategies as st
import pytest as pt
from ..monads import Maybe, Surely
from ..math import (
    SafeFloorDiv,
    SafeMod,
    SafeDiv,
    SafeExp,
    Undefined,
    SafeFactorial,
    SafeUnarySub,
    SafeUnaryAdd,
)
from math import inf, nan, isinf, isnan, factorial

MAX = 100000
MIN = -MAX


@given(
    x=st.integers(min_value=MIN, max_value=MAX) | st.floats(allow_nan=False),
    y=st.integers(min_value=MIN, max_value=MAX) | st.floats(allow_nan=False),
)
@hypothesis.settings(deadline=None)
@pt.mark.filterwarnings("ignore::RuntimeWarning")
def test_safe_div(x, y):
    # TODO SafeDiv/SafeExp/SafeMod: the specific edge case x=1 y=1 is slow
    assert SafeDiv(x, 0) == Undefined()
    assert SafeFloorDiv(x, 0) == Undefined()
    assume(x != 0 and y != 0)

    if isnan(x / y) and isnan(x // y):
        assert SafeDiv(x, y) == Undefined()
        assert SafeFloorDiv(x, y) == Undefined()
    else:
        assume(not isinf(x) and not isinf(y))
        assert SafeDiv(x, y) == x / y
        assert SafeFloorDiv(x, y) == x // y


@given(
    x=st.integers(min_value=MIN, max_value=MAX) | st.floats(allow_nan=False),
    y=st.integers(min_value=MIN, max_value=MAX) | st.floats(allow_nan=False),
)
@hypothesis.settings(deadline=None)
@pt.mark.filterwarnings("ignore::RuntimeWarning")
def test_safe_exp(x, y):
    assert SafeExp(0, 0) == Undefined()
    assert SafeExp(0, inf) == Undefined()
    assert SafeExp(inf, 0) == Undefined()
    assume(x != 0 and y != 0)
    try:
        x ** y
    except OverflowError:  # really big number
        assert SafeExp(x, y) == inf
    else:
        assert SafeExp(x, y) == x ** y


@given(
    x=st.integers(min_value=MIN, max_value=MAX) | st.floats(allow_nan=False),
    y=st.integers(min_value=MIN, max_value=MAX) | st.floats(allow_nan=False),
)
@hypothesis.settings(deadline=None)
@pt.mark.filterwarnings("ignore::RuntimeWarning")
def test_safe_mod(x, y):
    assert SafeMod(x, 0) == Undefined()
    assert SafeMod(inf, x) == Undefined()
    assume(x != 0 and y != 0)
    assume(not isinf(x))
    assert SafeMod(x, y) == x % y


@given(x=st.integers(min_value=0, max_value=MAX))
@hypothesis.settings(deadline=None)
@pt.mark.filterwarnings("ignore::RuntimeWarning")
def test_safe_factorial(x):
    assert SafeFactorial(inf) == Undefined()
    assert SafeFactorial(nan) == Undefined()
    assert SafeFactorial(x) == factorial(x)
    assume(x > 0)
    assert SafeFactorial(-x) == Undefined()


@given(x=st.integers(min_value=MIN, max_value=MAX))
@hypothesis.settings(deadline=None)
@pt.mark.filterwarnings("ignore::RuntimeWarning")
def test_safe_unary(x):
    assert SafeUnarySub(x) == -x
    assert SafeUnarySub(nan) == Undefined()
    assert SafeUnaryAdd(x) == +x
    assert SafeUnaryAdd(nan) == Undefined()


def test_undefined_sanity():
    assert Undefined() == Undefined()
    assert Undefined() + Undefined() == Undefined()
    assert Undefined() - Undefined() == Undefined()
    assert Undefined() * Undefined() == Undefined()
    assert Undefined() / Undefined() == Undefined()
    assert Undefined().__hash__() == hash(Undefined())
    assert Undefined().__nonzero__() is True
    assert Undefined().__index__() == 0
    assert Undefined().__oct__() == Undefined()
    # noinspection PyTypeChecker
    with pytest.raises(TypeError):
        from math import isnan

        assert isnan(complex(Undefined())) == isnan(complex(nan))
    try:
        from numpy import isnan
    except:
        pass
    else:
        assert isnan(complex(Undefined())) == isnan(complex(nan))
        del isnan
    from math import isnan

    assert isnan(float(Undefined())) == isnan(float(nan))
    assert Surely() == Surely()
    assert Surely() + Surely() == Surely()
    assert Surely() - Surely() == Surely()
    assert Surely() * Surely() == Surely()
    assert Surely() / Surely() == Surely()


@given(x=st.integers(min_value=MIN, max_value=MAX) | st.floats() | st.complex_numbers())
def test_number_undefined_sanity(x):
    assert Undefined() + x == Undefined()
    assert Undefined() - x == Undefined()
    assert Undefined() * x == Undefined()
    assert Undefined() / x == Undefined()
    assert Undefined() // x == Undefined()
    assert Surely() + x == Surely()
    assert Surely() - x == Surely()
    assert Surely() * x == Surely()
    assert Surely() / x == Surely()
    assert Undefined(x) != Surely(x)


@given(x=st.text() | st.characters())
def test_text_undefined_sanity(x):
    assert Undefined() + x == Undefined()
    assert Undefined() - x == Undefined()
    assert Undefined() * x == Undefined()
    assert Undefined() / x == Undefined()
    assert Undefined() // x == Undefined()
    assert Surely() + x == Surely()
    assert Surely() - x == Surely()
    assert Surely() * x == Surely()
    assert Surely() / x == Surely()
    assert Undefined(x) != Surely(x)


@given(x=st.lists(st.randoms()))
def test_list_undefined_sanity(x):
    assert Undefined() + x == Undefined()
    assert Undefined() - x == Undefined()
    assert Undefined() * x == Undefined()
    assert Undefined() / x == Undefined()
    assert Undefined() // x == Undefined()
    assert Surely() + x == Surely()
    assert Surely() - x == Surely()
    assert Surely() * x == Surely()
    assert Surely() / x == Surely()
    assert Undefined(x) != Surely(x)


@given(x=st.dictionaries(st.randoms(), st.randoms()))
def test_dict_undefined_sanity(x):
    assert Undefined() + x == Undefined()
    assert Undefined() - x == Undefined()
    assert Undefined() * x == Undefined()
    assert Undefined() / x == Undefined()
    assert Undefined() // x == Undefined()
    assert Surely() + x == Surely()
    assert Surely() - x == Surely()
    assert Surely() * x == Surely()
    assert Surely() / x == Surely()
    assert Undefined(x) != Surely(x)


@given(x=st.integers(min_value=MIN, max_value=MAX))
@hypothesis.settings(deadline=None)
def test_integer_monad_sanity(x):
    assert Maybe(x) != x
    assert Surely(x) == Surely(x)
    assert Maybe(x) != Surely(x)
    assert Surely(Maybe(x)) != x
    assert Surely(Undefined()) == Undefined()
    assert Surely(x) == x
    assert -Surely(x) == -x
    assert ~Surely(x) == ~x
    assert ~Maybe(x) != ~x
    assert -Maybe(x) != -x
    assert ~Surely() == Undefined()
    assert oct(Surely(x)) == oct(x)
    assert (not not Surely(x)) == bool(x)
    assert Surely(x) // 1 == x // 1
    assert Surely(x) / 1 == x / 1
    assert Surely(x) * 1 == x * 1
    assert ~Surely(x) == ~x
    assert -Surely(x) == -x
    assert Surely(x) - 1 == x - 1
    assert Surely(x) + 1 == x + 1
    assert (Surely(x) == 1) == (x == 1)
    assert hash(Surely(x)) == hash(x)



from .suggester import create_suggester
from .dummy import Any, Never
from .separate_context import SeparateContext
from .types_misc import Type, CheckType, Number
from .strings import *
from .lists import List
from .seq import Seq
from .tuple import Tuple
from .collection import Collection
from .datetime_tz import DatetimeWithTz
from .dicts import Dict
from .map import Map
from .sets import *
from .attributes import Attr
from .files import File


from .comparison import CheckOrder
from .arithmetic import Binary, Unary
from .compositions import OR, And, Not, composite_contract, or_contract
from .variables import (
    BindVariable,
    VariableRef,
    misc_variables_contract,
    int_variables_contract,
    misc_variables_ref,
    int_variables_ref,
)
from .simple_values import EqualTo, SimpleRValue

try:
    import numpy
except ImportError:  # pragma: no cover
    pass
else:
    from .array import (
        ShapeContract,
        Shape,
        Array,
        ArrayConstraint,
        DType,
        dtype,
        ArrayOR,
        ArrayAnd,
    )

from .extensions import (
    identifier_expression,
    identifier_contract,
    Extension,
    CheckCallable,
)

from .isinstance_imp import *
from .scoped_variables import scoped_variables_ref

from .miscellaneous_aliases import *



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec


def test_RestrictingNodeTransformer__test_ternary_if(mocker):
    src = "x.y = y.a if y.z else y.b"
    _getattr_ = mocker.stub()
    _getattr_.side_effect = lambda ob, key: ob[key]
    _write_ = mocker.stub()
    _write_.side_effect = lambda ob: ob

    glb = {
        "_getattr_": _getattr_,
        "_write_": _write_,
        "x": mocker.stub(),
        "y": {"a": "a", "b": "b"},
    }

    glb["y"]["z"] = True
    restricted_exec(src, glb)

    assert glb["x"].y == "a"
    _write_.assert_called_once_with(glb["x"])
    _getattr_.assert_has_calls([mocker.call(glb["y"], "z"), mocker.call(glb["y"], "a")])

    _write_.reset_mock()
    _getattr_.reset_mock()

    glb["y"]["z"] = False
    restricted_exec(src, glb)

    assert glb["x"].y == "b"
    _write_.assert_called_once_with(glb["x"])
    _getattr_.assert_has_calls([mocker.call(glb["y"], "z"), mocker.call(glb["y"], "b")])



from operator import getitem
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval


def test_slice():
    rglb = {"_getitem_": getitem}  # restricted globals

    assert restricted_eval("[1, 2, 3, 4, 5]", rglb) == [1, 2, 3, 4, 5]
    assert restricted_eval("[1, 2, 3, 4, 5][:]", rglb) == [1, 2, 3, 4, 5]
    assert restricted_eval("[1, 2, 3, 4, 5][1:]", rglb) == [2, 3, 4, 5]
    assert restricted_eval("[1, 2, 3, 4, 5][:4]", rglb) == [1, 2, 3, 4]
    assert restricted_eval("[1, 2, 3, 4, 5][1:4]", rglb) == [2, 3, 4]
    assert restricted_eval("[1, 2, 3, 4, 5][::3]", rglb) == [1, 4]
    assert restricted_eval("[1, 2, 3, 4, 5][1::3]", rglb) == [2, 5]
    assert restricted_eval("[1, 2, 3, 4, 5][:4:3]", rglb) == [1, 4]
    assert restricted_eval("[1, 2, 3, 4, 5][1:4:3]", rglb) == [2]



from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class CaptureTest(unittest.TestCase):
    def test_multi_capture(self):
        result = match(
            value=[
                {
                    "first-name": "Jane",
                    "last-name": "Doe",
                },
                {
                    "first-name": "John",
                    "last-name": "Doe",
                },
            ],
            pattern=Each({"first-name": Capture(..., name="first_names")}),
            multimatch=True,
        )
        self.assertTrue(result)
        self.assertEqual(["Jane", "John"], result["first_names"])

    def test_last_capture(self):
        result = match(
            multimatch=False,
            value=[
                {
                    "first-name": "Jane",
                    "last-name": "Doe",
                },
                {
                    "first-name": "John",
                    "last-name": "Doe",
                },
            ],
            pattern=Each({"first-name": Capture(..., name="first_name")}),
        )
        self.assertTrue(result)
        self.assertEqual("John", result["first_name"])

    def test_syntactic_sugar_rshift_operator(self):
        result = match(
            sample_k8s_response,
            {
                "containers": Each(
                    {
                        "image": _ >> "image",
                        "name": _ >> "name",
                        "ports": Each({"containerPort": _ >> "port"}),
                    }
                )
            },
        )
        self.assertTrue(result)
        self.assertEqual(
            "k8s.gcr.io/metrics-server/metrics-server:v0.4.1", result["image"]
        )
        self.assertEqual("metrics-server", result["name"])
        self.assertEqual(4443, result["port"])

    def test_syntactic_sugar_rmatmul_operator(self):
        # noinspection PyUnresolvedReferences
        result = match(
            sample_k8s_response,
            {
                "containers": Each(
                    {
                        "image": "image" @ _,
                        "name": "name" @ _,
                        "ports": Each({"containerPort": "port" @ _}),
                    }
                )
            },
        )
        self.assertTrue(result)
        self.assertEqual(
            "k8s.gcr.io/metrics-server/metrics-server:v0.4.1", result["image"]
        )
        self.assertEqual("metrics-server", result["name"])
        self.assertEqual(4443, result["port"])

    def test_syntactic_sugar_underscore(self):
        result = match(
            sample_k8s_response,
            {
                "containers": Each(
                    {
                        "image": _ >> "image",
                        "name": _ >> "name",
                        "ports": Each({"containerPort": _ >> "port"}),
                    }
                )
            },
        )
        self.assertTrue(result)
        self.assertEqual(
            "k8s.gcr.io/metrics-server/metrics-server:v0.4.1", result["image"]
        )
        self.assertEqual("metrics-server", result["name"])
        self.assertEqual(4443, result["port"])

    def test_regex_bind_groups(self):
        result = match(
            "abcdef", Regex(r"(?P<one>[a-c]+)(?P<two>[a-z]+)", bind_groups=True)
        )
        self.assertTrue(result)
        self.assertEqual(result["one"], "abc")
        self.assertEqual(result["two"], "def")
        self.assertEqual(result.one, "abc")
        self.assertEqual(result.two, "def")

    def test_nested_capture(self):
        result = match([1, 2, 3, 4], [1, Many(...) >> "x" >> "y", 4])
        self.assertTrue(result)
        self.assertEqual([2, 3], result.x)
        self.assertEqual([2, 3], result.y)


sample_k8s_response = {
    "containers": [
        {
            "args": [
                "--cert-dir=/tmp",
                "--secure-port=4443",
                "--kubelet-preferred-address-types=InternalIP,ExternalIP,Hostname",
                "--kubelet-use-node-status-port",
            ],
            "image": "k8s.gcr.io/metrics-server/metrics-server:v0.4.1",
            "imagePullPolicy": "IfNotPresent",
            "livenessProbe": {
                "failureThreshold": 3,
                "httpGet": {"path": "/livez", "port": "https", "scheme": "HTTPS"},
                "periodSeconds": 10,
                "successThreshold": 1,
                "timeoutSeconds": 1,
            },
            "name": "metrics-server",
            "ports": [{"containerPort": 4443, "name": "https", "protocol": "TCP"}],
            "readinessProbe": {
                "failureThreshold": 3,
                "httpGet": {"path": "/readyz", "port": "https", "scheme": "HTTPS"},
                "periodSeconds": 10,
                "successThreshold": 1,
                "timeoutSeconds": 1,
            },
            "resources": {},
            "securityContext": {
                "readOnlyRootFilesystem": True,
                "runAsNonRoot": True,
                "runAsUser": 1000,
            },
            "terminationMessagePath": "/dev/termination-log",
            "terminationMessagePolicy": "File",
            "volumeMounts": [
                {"mountPath": "/tmp", "name": "tmp-dir"},
                {
                    "mountPath": "/var/run/secrets/kubernetes.io/serviceaccount",
                    "name": "metrics-server-token-2j86f",
                    "readOnly": True,
                },
            ],
        }
    ],
    "dnsPolicy": "ClusterFirst",
    "enableServiceLinks": True,
    "nodeName": "ip-10-2-169-118.ec2.internal",
    "nodeSelector": {"kubernetes.io/os": "linux"},
    "priority": 2000000000,
    "priorityClassName": "system-cluster-critical",
    "restartPolicy": "Always",
    "schedulerName": "default-scheduler",
    "securityContext": {},
    "serviceAccount": "metrics-server",
    "serviceAccountName": "metrics-server",
    "terminationGracePeriodSeconds": 30,
    "tolerations": [
        {
            "effect": "NoExecute",
            "key": "node.kubernetes.io/not-ready",
            "operator": "Exists",
            "tolerationSeconds": 300,
        },
        {
            "effect": "NoExecute",
            "key": "node.kubernetes.io/unreachable",
            "operator": "Exists",
            "tolerationSeconds": 300,
        },
    ],
    "volumes": [
        {"emptyDir": {}, "name": "tmp-dir"},
        {
            "name": "metrics-server-token-2j86f",
            "secret": {"defaultMode": 420, "secretName": "metrics-server-token-2j86f"},
        },
    ],
}



from numpy import zeros, ones
import numpy as np

from . import syntax_fail, good, fail
from Aspidites._vendor.contracts.library.array import (
    np_int_dtypes,
    np_uint_dtypes,
    np_float_dtypes,
)

a_u8 = np.zeros((3, 4), dtype="uint8")
a_i8 = np.zeros((3, 4), dtype="int8")
a_u16 = np.zeros((3, 4), dtype="uint16")
a_i16 = np.zeros((3, 4), dtype="int16")
a_u32 = np.zeros((3, 4), dtype="uint32")
a_i32 = np.zeros((3, 4), dtype="int32")
a_f32 = np.zeros((3, 4), dtype="float32")
a_f64 = np.zeros((3, 4), dtype="float64")
a_bool = np.zeros((3, 4), dtype="bool")

# ## Strings
good("array", a_f32)
good("array", a_f64)
# synonims
good("ndarray", a_f32, exact=False)  # will be canonicalized to "array"
good("ndarray", a_f64, exact=False)  # same
fail("array", [0, 1])
# dtypes
good("array(uint8)", a_u8)
good("array(u1)", a_u8)
good("array(int8)", a_i8)
good("array(i1)", a_i8)
good("array(float32)", a_f32)
good("array(float64)", a_f64)
good("array(bool)", a_bool)
fail("array(float64)", a_f32)
fail("array(float32)", a_f64)
fail("array(uint8)", a_f32)
fail("array(float32)", a_u8)

# shapes
a0d = np.zeros((), dtype="float32")
a1d = np.zeros((2,))
a2d = np.zeros(
    (
        2,
        4,
    )
)
a3d = np.zeros((2, 4, 8))

good("shape[0]", a0d)
good("shape[1]", a1d)
good("shape[2]", a2d)
good("shape[3]", a3d)
fail("shape[>0]", a0d)
fail("shape[<1]", a1d)
fail("shape[>2]", a2d)
fail("shape[<3]", a3d)
good(["shape[x]", "shape[y],x=y"], [a3d, a3d])
good(["shape[x]", "shape[y],x=y"], [a2d, a2d])
good(["shape[x]", "shape[y],x=y"], [a1d, a1d])
good(["shape[x]", "shape[y],x=y"], [a0d, a0d])

good("array[2x4]", a2d)
fail("array[AxBxC]", a2d)
fail("array[2x4]", a3d)

good("array[HxW],H=2,W>3", a2d)
good("array[(=2)x(>3)]", a2d, exact=False)  # Parenthesis are unnecessary

# ellipsis to mean 0 or more dimensions
good("array[2x4x...]", a2d)
good("array[2x4x...]", a3d)

# if we really want more, use:
good("shape[>2],array[2x4x...]", a3d)
fail("shape[>2],array[2x4x...]", a2d)

fail("shape[>2]", [2, 2, 3])

# Try some binding:
good("array[XxYx...],X=2,Y=4", a3d)

# We don't do in between yet
syntax_fail("array[2x...x3]")

# We don't do in between yet
syntax_fail("array[2x...x3]")
# Try some binding:
good("array[XxYx...],X=2,Y=4", a2d)
good("array[XxYx...],X=2,Y=4", a3d)

# Using the array syntax  array[shape desc]

v1d = np.zeros(100)
v2d = np.zeros((10, 10))
good("array[100]", v1d)
fail("array[100]", v2d)
good("array[10x10]", v2d)
fail("array[10x...]", v1d)
good("array[10x...]", v2d)

fail("shape[>0]", a0d)
fail("shape[<1]", a1d)
fail("shape[>2]", a2d)
fail("shape[<3]", a3d)
good("shape[0]", a0d)
good("shape[1]", a1d)
good("shape[2]", a2d)
good("shape[3]", a3d)
good("array[shape[3]]", a3d)

good("array[1x2]", zeros((1, 2)))

# Now: special comparisons for arrays
A = np.array([0, 1, 2])
B = np.array([10, 20, 30])
good("array(>=0)", A)
good("array(<=2)", A)
fail("array(<=20)", B)
good("array(=0)", zeros((10)))
fail("array(=0)", np.array([0, 1, 0]))
good("array(=1)", ones((10)))

good(["shape(x)", "shape(x)"], [a2d, a2d])
fail(["shape(x)", "shape(x)"], [a2d, a3d])

good("array[NxN](<=1)", np.ones((10, 10)))
good("array[NxN](<=1,float32)", np.ones((10, 10), dtype="float32"))
good("array[NxN](<=1,float32|float64)", np.ones((10, 10), dtype="float64"))
good("array[NxN](<=1,(float32|float64))", np.ones((10, 10), dtype="float64"))
good("array[NxN](<=1,>=1)", np.ones((10, 10)))

# more complicated tests

good("array[(2|3)xN]", np.ones((2, 10)))
good("array[(2|3)xN]", np.ones((3, 10)))
fail("array[(2|3)xN]", np.ones((4, 10)))
good("array[(2|3)x...]", np.ones((2, 10)))
good("array[(2|3)x...]", np.ones((3, 10)))
fail("array[(2|3)x...]", np.ones((4, 10)))

good("array[(2,*)xN]", np.ones((2, 10)))
fail("array[(2,3)xN]", np.ones((4, 10)))
good("array[(2,*)x...]", np.ones((2, 10)))
fail("array[(2,3)x...]", np.ones((4, 10)))


good("seq", np.ones(3))
good("seq[3]", np.ones(3))
fail("seq[3]", np.ones(2))
good("seq[6]", np.ones((2, 3)))
fail("seq[6]", np.ones((2, 4)))

# "finite" contract
good("finite", 1)
good("finite", 0)
good("finite", -1)
good("finite", np.float_(1))
fail("finite", np.inf)
fail("finite", np.nan)

# some more complex stuff
good("array[N](>=-pi,<pi)", np.array([0], "float32"))
good("array[N](>=0)", np.array([0], "float32"))

# generalized ideas of numbers
# np_ints = ['int8', 'int16', 'int32', 'int64',
#            'uint8', 'uint16', 'uint32', 'uint64',
#            ]
# np_floats = ['float32', 'float64']

good("array(float)", np.array(1.32, "float32"))


for dt in np_int_dtypes:
    x = np.array(1).astype(dt)
    good("number", x)
    fail("Number", x)

    good("array(int)", x)
    good("int", x)
    fail("Int", x)

    fail("array(float)", x)
    fail("float", x)
    fail("Float", x)

# generalized ideas of numbers
for dt in np_float_dtypes:
    x = np.array(1).astype(dt)
    good("number", x)
    fail("Number", x)

    fail("array(int)", x)
    fail("int", x)
    fail("Int", x)

    good("array(float)", x)
    good("float", x)
    fail("Float", x)

# generalized ideas of numbers
for dt in np_uint_dtypes:
    x = np.array(1).astype(dt)
    good("number", x)
    fail("Number", x)

    fail("array(int)", x)
    fail("int", x)
    fail("Int", x)

    good("array(uint)", x)
    fail("Int", x)

    fail("array(float)", x)
    fail("float", x)
    fail("Float", x)



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from functools import reduce

from Aspidites._vendor._compat import basestring

from ..interface import RValue
from ..syntax import isnumber, W


class Binary(RValue):
    operations = {
        "+": lambda x, y: x + y,
        "-": lambda x, y: x - y,
        "*": lambda x, y: x * y,
        "^": lambda x, y: x ** y,
    }

    precedence = {
        "+": 0,
        "-": 0,
        "*": 1,
        "^": 2,
    }

    def __init__(self, exprs, glyph, where=None):
        assert glyph in Binary.operations
        for e in exprs:
            assert isinstance(e, RValue)

        self.where = where
        self.exprs = exprs
        self.glyph = glyph
        self.precedence = Binary.precedence[glyph]

    def eval(self, context):  # @ReservedAssignment
        vals = []
        for expr in self.exprs:
            val = expr.eval(context)
            if not isnumber(val):
                raise ValueError(
                    "I can only do math with numbers, not %r." % val.__class__.__name__
                )
            vals.append(val)
        operation = Binary.operations[self.glyph]
        return reduce(operation, vals)

    def __repr__(self):
        s = "Binary(%r,%r)" % (self.exprs, self.glyph)
        return s

    def __str__(self):
        def convert(x):
            if isinstance(x, Binary) and x.precedence < self.precedence:
                return "(%s)" % x
            else:
                return "%s" % x

        s = self.glyph.join(convert(x) for x in self.exprs)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        l = list(tokens[0])
        exprs = [l.pop(0)]
        while l:
            glyph = l.pop(0)
            assert isinstance(glyph, basestring)
            expr = l.pop(0)
            assert isinstance(expr, RValue)
            exprs.append(expr)

        # noinspection PyUnboundLocalVariable
        return Binary(exprs, glyph, where=where)


class Unary(RValue):

    operations = {
        "-": lambda x: (-x),
    }

    def __init__(self, glyph, expr, where=None):
        assert glyph in Unary.operations
        assert isinstance(expr, RValue)

        self.where = where
        self.expr = expr
        self.glyph = glyph

    def eval(self, context):  # @ReservedAssignment
        val = self.expr.eval(context)
        if not isnumber(val):
            raise ValueError(
                "I can only do math with numbers, not with %r." % val.__class__.__name__
            )

        operation = Unary.operations[self.glyph]
        return operation(val)

    def __repr__(self):
        s = "Unary(%r,%r)" % (self.glyph, self.expr)
        return s

    def __str__(self):
        # XXX: precedence
        return "%s%s" % (self.glyph, self.expr)

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        glyph = tokens[0][0]
        assert isinstance(glyph, basestring)
        expr = tokens[0][1]
        assert isinstance(expr, RValue)
        return Unary(glyph, expr, where=where)



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval


def test_Eq():
    assert restricted_eval("1 == 1") is True


def test_NotEq():
    assert restricted_eval("1 != 2") is True


def test_Gt():
    assert restricted_eval("2 > 1") is True


def test_Lt():
    assert restricted_eval("1 < 2")


def test_GtE():
    assert restricted_eval("2 >= 2") is True


def test_LtE():
    assert restricted_eval("1 <= 2") is True



from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class TryMatchTest(unittest.TestCase):
    def test_try_match(self):
        result = None
        id_ = None
        try:
            match({"id": "h23ksad8", "name": "ObjectName"})
        except Case({"id": Capture(Regex("[A-Z0-9]+"), name="id")}):
            result = "upper"
        except Case({"id": Capture(Regex("[a-z0-9]+"), name="id")}) as match_result:
            result = "lower"
            id_ = match_result["id"]
        finally:
            self.assertEqual("lower", result)
            self.assertEqual("h23ksad8", id_)

    def test_try_match_default(self):
        result = None
        try:
            match({"id": "h23ksad8", "name": "ObjectName"})
        except Case({"id": Regex("[A-Z0-9]+")}):
            result = "upper"
        except Default:
            result = "unknown"
        finally:
            self.assertEqual(result, "unknown")



__version__ = "2.0.1"

from .interface import (
    Contract,
    ContractNotRespected,
    CannotDecorateClassmethods,
    ContractSyntaxError,
    ContractException,
)

from .main import (
    contract_decorator,
    check,
    fail,
    check_multiple,
    contracts_decorate as decorate,
    parse_flexible_spec as parse,
)


# Just make them appear as belonging to the "contracts" Module
# So that Eclipse and other IDEs will not get confused.
contract = contract_decorator  # type: ignore

contract.__doc__ = contract_decorator.__doc__

from .main import new_contract as new_contract_main


def new_contract(*args):
    return new_contract_main(*args)


new_contract.__doc__ = new_contract_main.__doc__

# A couple of useful functions
from .interface import describe_value, describe_type, describe_value_multiline
from .utils import *

from .metaclass import ContractsMeta, with_metaclass

ContractsMeta.__module__ = "contracts"

# And after everything else is loaded, load the  utils
from .useful_contracts import *

# After everything is loaded, load aliases
from .library import miscellaneous_aliases  # @UnusedImport



from . import syntax_fail, good, fail


#### Tuples
good("tuple", ())
good("tuple", (1,))
# tuples and lists are different
fail("tuple", [])
fail("list", ())
# tuples can have the length
good("tuple[*]", (2, 2))
good("tuple[1]", (1,))
# you cannot specify every element
good("tuple(*,*)", (1, 2))
good("tuple(*)", (1,))
fail("tuple(*,*)", (1, 2, 3))
good("tuple(int,int)", (1, 2))
good("tuple(int,float)", (1, 2.0))
fail("tuple(float,float)", (1, 2.0))
good("tuple(type(x),type(x))", (1, 2))
# something complicated: nested tuples
good("tuple(x,tuple(*,*,x))", (1, (2, 3, 1)))
fail("tuple(x,tuple(*,*,x))", (1, (2, 3, 2)))
good("tuple(type(x),tuple(*,*,type(x)))", (1, (2.1, 3.0, 3)))
fail("tuple(type(x),tuple(*,*,type(x)))", (1, (2.1, 3.0, 3.1)))
# cannot specify both, even if coherent
syntax_fail("tuple[*](*,*)")

good("tuple(0,1|2)", (0, 2))
good("tuple(0,1|2)", (0, 2))
good("tuple(0,1|2|3)", (0, 3))
good("tuple(0,1|2|3,4)", (0, 3, 4))
fail("tuple(0,1|2)", (0, 3))
good("tuple(0,1,2)", (0, 1, 2))
good("tuple(1|2,3)", (1, 3))
good("tuple(1,(>2,int))", (1, 3))
fail("tuple(1,(>2,int))", (1, 3.0))
good("tuple(1,(*,*),2)", (1, 3, 2))
good("tuple(str,(str[1],str))", ("a", "b"))



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from numbers import Number
import math

# All the imports from pyparsing go here
from typing import Any, List

from types import ModuleType

from Aspidites._vendor.pyparsing import (
    delimitedList,
    Forward,
    Literal,
    stringEnd,
    nums,
    Word,
    CaselessLiteral,
    Combine,
    Optional,
    Suppress,
    OneOrMore,
    ZeroOrMore,
    opAssoc,
    infixNotation as operatorPrecedence,
    oneOf,
    ParseException,
    ParserElement,
    alphas,
    alphanums,
    ParseFatalException,
    ParseSyntaxException,
    FollowedBy,
    NotAny,
    Or,
    MatchFirst,
    Keyword,
    Group,
    White,
    lineno,
    col,
)

# from .pyparsing_utils import myOperatorPrecedence


# Enable memoization (much faster!)
ParserElement.enablePackrat(cache_size_limit=None)
# else:
#     # Pyparsing 2.0
#     from pyparsing import infixNotation
#     myOperatorPrecendence = infixNotation

from .interface import Where


class ParsingTmp:
    contract_types: List[Any] = []
    keywords: List[Any] = []


def add_contract(x):
    ParsingTmp.contract_types.append(x)


def add_keyword(x):
    """Declares that x is a keyword --- this is useful to have more
    clear messages. "keywords" are not parsed by Extension.
    (see extensions.py) and allows to have "deep" error indications.
    See http://pyparsing.wikispaces.com/message/view/home/620225
    and the discussion of the "-" operator in the docs.
    """
    ParsingTmp.keywords.append(x)


W = Where

O = Optional
S = Suppress

basenumber = Word(nums)
point = Literal(".")
e = CaselessLiteral("E")
plusorminus = Literal("+") | Literal("-")
integer = Combine(O(plusorminus) + basenumber)
integer.setParseAction(lambda x, y, tokens: SimpleRValue(int(tokens[0])))
floatnumber = Combine(
    O(plusorminus) + integer + (point + O(basenumber)) ^ (e + integer)
)
floatnumber.setParseAction(lambda x, y, tokens: SimpleRValue(float(tokens[0])))
pi = Keyword("pi").setParseAction(
    lambda x, y, tokens: SimpleRValue(math.pi, "pi")
)  # @UnusedVariable

np_ = False

try:
    import numpy
except ImportError:
    pass
else:
    np_ = True


def isnumber(x):
    # These are scalar quantities that we can compare (=,>,>=, etc.)
    if isinstance(x, Number):
        return True

    if np_ and isinstance(x, numpy.number):
        return True

    return False


rvalue = Forward()
rvalue.setName("rvalue")
contract_expression = Forward()
contract_expression.setName("contract")
simple_contract = Forward()
simple_contract.setName("simple_contract")

# Import all expressions -- they will call add_contract()
from .library import (
    EqualTo,
    Unary,
    Binary,
    composite_contract,
    identifier_contract,
    misc_variables_contract,
    scoped_variables_ref,
    int_variables_contract,
    int_variables_ref,
    misc_variables_ref,
    SimpleRValue,
)

number = pi | floatnumber | integer
operand = number | int_variables_ref | misc_variables_ref | scoped_variables_ref
operand.setName("r-value")

op = operatorPrecedence
# op  = myOperatorPrecedence
rvalue << op(
    operand,
    [
        ("-", 1, opAssoc.RIGHT, Unary.parse_action),
        ("*", 2, opAssoc.LEFT, Binary.parse_action),
        ("-", 2, opAssoc.LEFT, Binary.parse_action),
        ("+", 2, opAssoc.LEFT, Binary.parse_action),
        ("^", 2, opAssoc.LEFT, Binary.parse_action),
    ],
)

# I want
# - BindVariable to have precedence to EqualTo(VariableRef)
# but I also want:
# - Arithmetic to have precedence w.r.t BindVariable
# last is variables
add_contract(misc_variables_contract)
add_contract(int_variables_contract)

add_contract(rvalue.copy().setParseAction(EqualTo.parse_action))

hardwired = MatchFirst(ParsingTmp.contract_types)
hardwired.setName("Predefined contract expression")

simple_contract << (hardwired | identifier_contract)
simple_contract.setName("simple contract expression")

any_contract = composite_contract | simple_contract
any_contract.setName("Any simple or composite contract")
contract_expression << (any_contract)  # Parentheses before << !!



from .syntax import Forward, Suppress, FollowedBy, Group, OneOrMore, Optional, opAssoc


def myOperatorPrecedence(baseExpr, opList):
    """Helper method for constructing grammars of expressions made up of
    operators working in a precedence hierarchy.  Operators may be unary or
    binary, left- or right-associative.  Parse actions can also be attached
    to operator expressions.

    Parameters:
     - baseExpr - expression representing the most basic element for the
     - opList - list of tuples, one for each operator precedence level in
       expression grammar; each tuple is of the form
       (opExpr, numTerms, rightLeftAssoc, parseAction), where:
        - opExpr is the pyparsing expression for the operator;
           may also be a string, which will be converted to a Literal;
           if numTerms is 3, opExpr is a tuple of two expressions, for the
           two operators separating the 3 terms
        - numTerms is the number of terms for this operator (must
           be 1, 2, or 3)
        - rightLeftAssoc is the indicator whether the operator is
           right or left associative, using the pyparsing-defined
           constants opAssoc.RIGHT and opAssoc.LEFT.
        - parseAction is the parse action to be associated with
           expressions matching this operator expression (the
           parse action tuple member may be omitted)
    """
    ret = Forward()
    allops = [x[0] for x in opList]
    opnames = ",".join(str(x) for x in allops)
    ret.setName("operatorSystem(%s)" % opnames)
    # parenthesis = Suppress('(') + ret + FollowedBy(NotAny(oneOf(allops)))
    # - Suppress(')')
    parenthesis = Suppress("(") - ret - Suppress(")")
    lastExpr = parenthesis.setName("parenthesis(%s)" % opnames) | baseExpr
    # lastExpr.setName('Base operand (%s) or parenthesis' % str(baseExpr))
    for operDef in opList:
        opExpr, arity, rightLeftAssoc, pa = (operDef + (None,))[:4]
        if arity == 3:
            if opExpr is None or len(opExpr) != 2:
                raise ValueError(
                    "if numterms=3, opExpr must be a tuple or list of two "
                    "expressions"
                )
            opExpr1, opExpr2 = opExpr
        thisExpr = Forward().setName("operation_with(%s)" % opExpr)
        if rightLeftAssoc == opAssoc.LEFT:
            if arity == 1:
                matchExpr = FollowedBy(lastExpr + opExpr) + Group(
                    lastExpr + OneOrMore(opExpr)
                )
            elif arity == 2:
                if opExpr is not None:
                    #                    matchExpr = Group(lastExpr +
                    # FollowedBy(opExpr) + OneOrMore(opExpr - lastExpr))

                    matchExpr = Group(
                        lastExpr + FollowedBy(opExpr) - OneOrMore(opExpr - lastExpr)
                    )
                else:
                    matchExpr = FollowedBy(lastExpr + lastExpr) + Group(
                        lastExpr + OneOrMore(lastExpr)
                    )
            elif arity == 3:
                matchExpr = FollowedBy(
                    lastExpr + opExpr1 + lastExpr + opExpr2 + lastExpr
                ) + Group(lastExpr + opExpr1 + lastExpr + opExpr2 + lastExpr)
            else:
                raise ValueError(
                    "operator must be unary (1), binary (2), or ternary (3)"
                )
        elif rightLeftAssoc == opAssoc.RIGHT:
            if arity == 1:
                # try to avoid LR with this extra test
                if not isinstance(opExpr, Optional):
                    opExpr = Optional(opExpr)
                matchExpr = FollowedBy(opExpr.expr - thisExpr) - Group(
                    opExpr - thisExpr
                )
            elif arity == 2:
                if opExpr is not None:
                    matchExpr = FollowedBy(lastExpr + opExpr - thisExpr) + Group(
                        lastExpr + OneOrMore(opExpr - thisExpr)
                    )
                else:
                    matchExpr = FollowedBy(lastExpr + thisExpr) + Group(
                        lastExpr + OneOrMore(thisExpr)
                    )
            elif arity == 3:
                matchExpr = FollowedBy(
                    lastExpr + opExpr1 + thisExpr + opExpr2 + thisExpr
                ) + Group(lastExpr + opExpr1 + thisExpr + opExpr2 + thisExpr)
            else:
                raise ValueError(
                    "operator must be unary (1), binary (2), or ternary (3)"
                )
        else:
            raise ValueError("operator must indicate right or left associativity")
        if pa:
            matchExpr.setParseAction(pa)
        thisExpr << (matchExpr | lastExpr).setName("operation with %s" % opExpr)
        lastExpr = thisExpr
    ret << lastExpr
    return ret



from Aspidites._vendor.RestrictedPython.Utilities import reorder
from Aspidites._vendor.RestrictedPython.Utilities import test
import math
import random
import string


def test_string_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["string"] is string


def test_math_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["math"] is math


def test_whrandom_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["whrandom"] is random


def test_random_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["random"] is random


def test_set_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["set"] is set


def test_frozenset_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["frozenset"] is frozenset


def test_DateTime_in_utility_builtins_if_importable():
    try:
        import DateTime  # type: ignore
    except ImportError:
        pass
    else:
        from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

        assert DateTime.__name__ in utility_builtins


def test_same_type_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import same_type
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["same_type"] is same_type


def test_test_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import test
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["test"] is test


def test_reorder_in_utility_builtins():
    from Aspidites._vendor.RestrictedPython.Utilities import reorder
    from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins

    assert utility_builtins["reorder"] is reorder


def test_sametype_only_one_arg():
    from Aspidites._vendor.RestrictedPython.Utilities import same_type

    assert same_type(object())


def test_sametype_only_two_args_same():
    from Aspidites._vendor.RestrictedPython.Utilities import same_type

    assert same_type(object(), object())


def test_sametype_only_two_args_different():
    from Aspidites._vendor.RestrictedPython.Utilities import same_type

    class Foo(object):
        pass

    assert same_type(object(), Foo()) is False


def test_sametype_only_multiple_args_same():
    from Aspidites._vendor.RestrictedPython.Utilities import same_type

    assert same_type(object(), object(), object(), object())


def test_sametype_only_multipe_args_one_different():
    from Aspidites._vendor.RestrictedPython.Utilities import same_type

    class Foo(object):
        pass

    assert same_type(object(), object(), Foo()) is False


def test_test_single_value_true():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(True) is True


def test_test_single_value_False():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(False) is False


def test_test_even_values_first_true():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(True, "first", True, "second") == "first"


def test_test_even_values_not_first_true():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(False, "first", True, "second") == "second"


def test_test_odd_values_first_true():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(True, "first", True, "second", False) == "first"


def test_test_odd_values_not_first_true():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(False, "first", True, "second", False) == "second"


def test_test_odd_values_last_true():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(False, "first", False, "second", "third") == "third"


def test_test_odd_values_last_false():
    from Aspidites._vendor.RestrictedPython.Utilities import test

    assert test(False, "first", False, "second", False) is False


def test_reorder_with__None():
    from Aspidites._vendor.RestrictedPython.Utilities import reorder

    before = ["a", "b", "c", "d", "e"]
    without = ["a", "c", "e"]
    after = reorder(before, without=without)
    assert after == [("b", "b"), ("d", "d")]


def test_reorder_with__not_None():
    from Aspidites._vendor.RestrictedPython.Utilities import reorder

    before = ["a", "b", "c", "d", "e"]
    with_ = ["a", "d"]
    without = ["a", "c", "e"]
    after = reorder(before, with_=with_, without=without)
    assert after == [("d", "d")]


def test_Utilities__test_1():
    """It returns the first arg after the first argument which is True"""
    assert test(True, 1, False, 2) == 1
    assert test(False, 1, True, 2) == 2
    assert test(False, 1, False, 2, True, 3) == 3


def test_Utilities__test_2():
    """If the above is not met, and there is an extra argument
    it returns it."""
    assert test(False, 1, False, 2, 3) == 3
    assert test(False, 1, 2) == 2
    assert test(1) == 1
    assert not test(False)


def test_Utilities__test_3():
    """It returns None if there are only False args followed by something."""
    assert test(False, 1) is None
    assert test(False, 1, False, 2) is None


def test_Utilities__reorder_1():
    """It also supports 2-tuples containing key, value."""
    s = [("k1", "v1"), ("k2", "v2"), ("k3", "v3")]
    _with = [("k2", "v2"), ("k3", "v3")]
    without = [("k2", "v2"), ("k4", "v4")]
    assert reorder(s, _with, without) == [("k3", "v3")]



# fourFn.py
#
# Demonstration of the pyparsing module, implementing a simple 4-function expression parser,
# with support for scientific notation, and symbols for e and pi.
# Extended to add exponentiation and simple built-in functions.
# Extended test cases, simplified pushFirst method.
# Removed unnecessary expr.suppress() call (thanks Nathaniel Peterson!), and added Group
# Changed fnumber to use a Regex, which is now the preferred method
#
# Copyright 2003-2009 by Paul McGuire
#
from ....._vendor.pyparsing import (
    Literal,
    Word,
    Group,
    ZeroOrMore,
    Forward,
    alphas,
    alphanums,
    Regex,
    ParseException,
    CaselessKeyword,
    Suppress,
)
import math
import operator

exprStack = []


def pushFirst(strg, loc, toks):
    exprStack.append(toks[0])


def pushUMinus(strg, loc, toks):
    for t in toks:
        if t == "-":
            exprStack.append("unary -")
            # ~ exprStack.append( '-1' )
            # ~ exprStack.append( '*' )
        else:
            break


bnf = None


def BNF():
    """
    expop   :: '^'
    multop  :: '*' | '/'
    addop   :: '+' | '-'
    integer :: ['+' | '-'] '0'..'9'+
    atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
    factor  :: atom [ expop factor ]*
    term    :: factor [ multop factor ]*
    expr    :: term [ addop term ]*
    """
    global bnf
    if not bnf:
        point = Literal(".")
        # use CaselessKeyword for e and pi, to avoid accidentally matching
        # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
        # and CaselessKeyword only match whole words
        e = CaselessKeyword("E")
        pi = CaselessKeyword("PI")
        # ~ fnumber = Combine( Word( "+-"+nums, nums ) +
        # ~ Optional( point + Optional( Word( nums ) ) ) +
        # ~ Optional( e + Word( "+-"+nums, nums ) ) )
        fnumber = Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
        ident = Word(alphas, alphanums + "_$")

        plus, minus, mult, div = map(Literal, "+-*/")
        lpar, rpar = map(Suppress, "()")
        addop = plus | minus
        multop = mult | div
        expop = Literal("^")

        expr = Forward()
        atom = (
            (0, None) * minus
            + (pi | e | fnumber | ident + lpar + expr + rpar | ident).setParseAction(
                pushFirst
            )
            | Group(lpar + expr + rpar)
        ).setParseAction(pushUMinus)

        # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left exponents, instead of left-to-righ
        # that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor << atom + ZeroOrMore((expop + factor).setParseAction(pushFirst))

        term = factor + ZeroOrMore((multop + factor).setParseAction(pushFirst))
        expr << term + ZeroOrMore((addop + term).setParseAction(pushFirst))
        bnf = expr
    return bnf


# map operator symbols to corresponding arithmetic operations
epsilon = 1e-12
opn = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "^": operator.pow,
}
fn = {
    "sin": math.sin,
    "cos": math.cos,
    "tan": math.tan,
    "exp": math.exp,
    "abs": abs,
    "trunc": lambda a: int(a),
    "round": round,
    "sgn": lambda a: (a > epsilon) - (a < -epsilon),
}


def evaluateStack(s):
    op = s.pop()
    if op == "unary -":
        return -evaluateStack(s)
    if op in "+-*/^":
        op2 = evaluateStack(s)
        op1 = evaluateStack(s)
        return opn[op](op1, op2)
    elif op == "PI":
        return math.pi  # 3.1415926535
    elif op == "E":
        return math.e  # 2.718281828
    elif op in fn:
        return fn[op](evaluateStack(s))
    elif op[0].isalpha():
        raise Exception("invalid identifier '%s'" % op)
    else:
        return float(op)


if __name__ == "__main__":

    def test(s, expVal):
        global exprStack
        exprStack[:] = []
        try:
            results = BNF().parseString(s, parseAll=True)
            val = evaluateStack(exprStack[:])
        except ParseException as pe:
            print(s, "failed parse:", str(pe))
        except Exception as e:
            print(s, "failed eval:", str(e))
        else:
            if val == expVal:
                print(s, "=", val, results, "=>", exprStack)
            else:
                print(s + "!!!", val, "!=", expVal, results, "=>", exprStack)

    test("9", 9)
    test("-9", -9)
    test("--9", 9)
    test("-E", -math.e)
    test("9 + 3 + 6", 9 + 3 + 6)
    test("9 + 3 / 11", 9 + 3.0 / 11)
    test("(9 + 3)", (9 + 3))
    test("(9+3) / 11", (9 + 3.0) / 11)
    test("9 - 12 - 6", 9 - 12 - 6)
    test("9 - (12 - 6)", 9 - (12 - 6))
    test("2*3.14159", 2 * 3.14159)
    test("3.1415926535*3.1415926535 / 10", 3.1415926535 * 3.1415926535 / 10)
    test("PI * PI / 10", math.pi * math.pi / 10)
    test("PI*PI/10", math.pi * math.pi / 10)
    test("PI^2", math.pi ** 2)
    test("round(PI^2)", round(math.pi ** 2))
    test("6.02E23 * 8.048", 6.02e23 * 8.048)
    test("e / 3", math.e / 3)
    test("sin(PI/2)", math.sin(math.pi / 2))
    test("trunc(E)", int(math.e))
    test("trunc(-E)", int(-math.e))
    test("round(E)", round(math.e))
    test("round(-E)", round(-math.e))
    test("E^PI", math.e ** math.pi)
    test("exp(0)", 1)
    test("exp(1)", math.e)
    test("2^3^2", 2 ** 3 ** 2)
    test("2^3+2", 2 ** 3 + 2)
    test("2^3+5", 2 ** 3 + 5)
    test("2^9", 2 ** 9)
    test("sgn(-2)", -1)
    test("sgn(0)", 0)
    test("foo(0.1)", None)
    test("sgn(0.1)", 1)


"""
Test output:
>pythonw -u fourFn.py
9 = 9.0 ['9'] => ['9']
9 + 3 + 6 = 18.0 ['9', '+', '3', '+', '6'] => ['9', '3', '+', '6', '+']
9 + 3 / 11 = 9.27272727273 ['9', '+', '3', '/', '11'] => ['9', '3', '11', '/', '+']
(9 + 3) = 12.0 [] => ['9', '3', '+']
(9+3) / 11 = 1.09090909091 ['/', '11'] => ['9', '3', '+', '11', '/']
9 - 12 - 6 = -9.0 ['9', '-', '12', '-', '6'] => ['9', '12', '-', '6', '-']
9 - (12 - 6) = 3.0 ['9', '-'] => ['9', '12', '6', '-', '-']
2*3.14159 = 6.28318 ['2', '*', '3.14159'] => ['2', '3.14159', '*']
3.1415926535*3.1415926535 / 10 = 0.986960440053 ['3.1415926535', '*', '3.1415926535', '/', '10'] => ['3.1415926535', '3.1415926535', '*', '10', '/']
PI * PI / 10 = 0.986960440109 ['PI', '*', 'PI', '/', '10'] => ['PI', 'PI', '*', '10', '/']
PI*PI/10 = 0.986960440109 ['PI', '*', 'PI', '/', '10'] => ['PI', 'PI', '*', '10', '/']
PI^2 = 9.86960440109 ['PI', '^', '2'] => ['PI', '2', '^']
6.02E23 * 8.048 = 4.844896e+024 ['6.02E23', '*', '8.048'] => ['6.02E23', '8.048', '*']
e / 3 = 0.90609394282 ['E', '/', '3'] => ['E', '3', '/']
sin(PI/2) = 1.0 ['sin', 'PI', '/', '2'] => ['PI', '2', '/', 'sin']
trunc(E) = 2 ['trunc', 'E'] => ['E', 'trunc']
E^PI = 23.1406926328 ['E', '^', 'PI'] => ['E', 'PI', '^']
2^3^2 = 512.0 ['2', '^', '3', '^', '2'] => ['2', '3', '2', '^', '^']
2^3+2 = 10.0 ['2', '^', '3', '+', '2'] => ['2', '3', '^', '2', '+']
2^9 = 512.0 ['2', '^', '9'] => ['2', '9', '^']
sgn(-2) = -1 ['sgn', '-2'] => ['-2', 'sgn']
sgn(0) = 0 ['sgn', '0'] => ['0', 'sgn']
sgn(0.1) = 1 ['sgn', '0.1'] => ['0.1', 'sgn']
>Exit code: 0
"""



import builtins
import collections
import datetime
import functools
import importlib
import Aspidites._vendor.contracts.inspection as inspect
import io
import linecache
import os
from os.path import normcase
import _pickle
import pickle
import shutil
import sys
import types
import textwrap
import unicodedata
import unittest
import unittest.mock
import warnings

try:
    from concurrent.futures import ThreadPoolExecutor
except ImportError:
    ThreadPoolExecutor = None

from Aspidites.tests.support import cpython_only
from Aspidites.tests.support import MISSING_C_DOCSTRINGS, ALWAYS_EQ

# from Aspidites.tests.support.os_helper import TESTFN
from Aspidites.tests.support.script_helper import (
    assert_python_ok,
    assert_python_failure,
)
from Aspidites.tests import inspect_fodder as mod

# from Aspidites.tests import inspect_fodder2 as mod2
# from Aspidites.tests import support
# from Aspidites.tests import inspect_stock_annotations
# from Aspidites.tests import inspect_stringized_annotations
# from Aspidites.tests import inspect_stringized_annotations_2

# from Aspidites.tests.test_import import _ready_to_import

modfile = mod.__file__
if modfile.endswith(("c", "o")):
    modfile = modfile[:-1]

modfile = normcase(modfile)


def revise(filename, *args):
    return (normcase(filename),) + args


git = mod.StupidGit()


def signatures_with_lexicographic_keyword_only_parameters():
    """
    Yields a whole bunch of functions with only keyword-only parameters,
    where those parameters are always in lexicographically sorted order.
    """
    parameters = ["a", "bar", "c", "delta", "ephraim", "magical", "yoyo", "z"]
    for i in range(1, 2 ** len(parameters)):
        p = []
        bit = 1
        for j in range(len(parameters)):
            if i & (bit << j):
                p.append(parameters[j])
        fn_text = "def foo(*, " + ", ".join(p) + "): pass"
        symbols = {}
        exec(fn_text, symbols, symbols)
        yield symbols["foo"]


def unsorted_keyword_only_parameters_fn(
    *, throw, out, the, baby, with_, the_, bathwater
):
    pass


unsorted_keyword_only_parameters = "throw out the baby with_ the_ bathwater".split()


class IsTestBase(unittest.TestCase):
    predicates = {
        inspect.isclass,
        inspect.iscode,
        inspect.isframe,
        inspect.isfunction,
        inspect.ismethod,
        inspect.ismodule,
        inspect.istraceback,
    }

    def istest(self, predicate, exp):
        obj = eval(exp)
        self.assertTrue(predicate(obj), "%s(%s)" % (predicate.__name__, exp))

        for other in self.predicates - set([predicate]):
            if other == inspect.isfunction:
                continue
            self.assertFalse(other(obj), "not %s(%s)" % (other.__name__, exp))


def generator_function_example(self):
    for i in range(2):
        yield i


async def async_generator_function_example(self):
    async for i in range(2):
        yield i


async def coroutine_function_example(self):
    return "spam"


@types.coroutine
def gen_coroutine_function_example(self):
    yield
    return "spam"


class TestPredicates(IsTestBase):
    def test_excluding_predicates(self):
        global tb

        self.istest(inspect.iscode, "mod.spam.__code__")
        try:
            1 / 0
        except:
            tb = sys.exc_info()[2]
            self.istest(inspect.isframe, "tb.tb_frame")
            self.istest(inspect.istraceback, "tb")

        finally:

            tb = None
        self.istest(inspect.isfunction, "mod.spam")
        self.istest(inspect.isfunction, "mod.StupidGit.abuse")
        self.istest(inspect.ismethod, "git.argue")
        self.istest(inspect.ismethod, "mod.custom_method")
        self.istest(inspect.ismodule, "mod")

    def test_isclass(self):
        self.istest(inspect.isclass, "mod.StupidGit")
        self.assertTrue(inspect.isclass(list))

        class CustomGetattr(object):
            def __getattr__(self, attr):
                return None

        self.assertFalse(inspect.isclass(CustomGetattr()))

    def test_get_slot_members(self):
        class C(object):
            __slots__ = ("a", "b")

        x = C()
        x.a = 42

    def test_isabstract(self):
        from abc import ABCMeta, abstractmethod

        class AbstractClassExample(metaclass=ABCMeta):
            @abstractmethod
            def foo(self):
                pass

        class ClassExample(AbstractClassExample):
            def foo(self):
                pass

        a = ClassExample()


class SlotUser:
    "Docstrings for __slots__"
    __slots__ = {"power": "measured in kilowatts", "distance": "measured in kilometers"}


class _BrokenDataDescriptor(object):
    """
    A broken data descriptor. See bug
    """

    def __get__(*args):
        raise AttributeError("broken data descriptor")

    def __set__(*args):
        raise RuntimeError

    def __getattr__(*args):
        raise AttributeError("broken data descriptor")


class _BrokenMethodDescriptor(object):
    """
    A broken method descriptor. See bug
    """

    def __get__(*args):
        raise AttributeError("broken method descriptor")

    def __getattr__(*args):
        raise AttributeError("broken method descriptor")


def attrs_wo_objs(cls):
    return [t[:3] for t in inspect.classify_class_attrs(cls)]


_global_ref = object()


class MyParameter(inspect.Parameter):
    pass


class TestParameterObject(unittest.TestCase):
    def test_signature_parameter_kinds(self):
        P = inspect.Parameter
        self.assertTrue(
            P.POSITIONAL_ONLY
            < P.POSITIONAL_OR_KEYWORD
            < P.VAR_POSITIONAL
            < P.KEYWORD_ONLY
            < P.VAR_KEYWORD
        )

        self.assertEqual(str(P.POSITIONAL_ONLY), "POSITIONAL_ONLY")
        self.assertTrue("POSITIONAL_ONLY" in repr(P.POSITIONAL_ONLY))

    def test_signature_parameter_object(self):
        p = inspect.Parameter("foo", default=10, kind=inspect.Parameter.POSITIONAL_ONLY)
        self.assertEqual(p.name, "foo")
        self.assertEqual(p.default, 10)
        self.assertIs(p.annotation, p.empty)
        self.assertEqual(p.kind, inspect.Parameter.POSITIONAL_ONLY)

        with self.assertRaisesRegex(
            ValueError, "value '123' is " "not a valid Parameter.kind"
        ):
            inspect.Parameter("foo", default=10, kind="123")

        with self.assertRaisesRegex(ValueError, "not a valid parameter name"):
            inspect.Parameter("1", kind=inspect.Parameter.VAR_KEYWORD)

        with self.assertRaisesRegex(TypeError, "name must be a str"):
            inspect.Parameter(None, kind=inspect.Parameter.VAR_KEYWORD)

        with self.assertRaisesRegex(ValueError, "is not a valid parameter name"):
            inspect.Parameter("$", kind=inspect.Parameter.VAR_KEYWORD)

        with self.assertRaisesRegex(ValueError, "is not a valid parameter name"):
            inspect.Parameter(".a", kind=inspect.Parameter.VAR_KEYWORD)

        with self.assertRaisesRegex(ValueError, "cannot have default values"):
            inspect.Parameter("a", default=42, kind=inspect.Parameter.VAR_KEYWORD)

        with self.assertRaisesRegex(ValueError, "cannot have default values"):
            inspect.Parameter("a", default=42, kind=inspect.Parameter.VAR_POSITIONAL)

        p = inspect.Parameter(
            "a", default=42, kind=inspect.Parameter.POSITIONAL_OR_KEYWORD
        )
        with self.assertRaisesRegex(ValueError, "cannot have default values"):
            p.replace(kind=inspect.Parameter.VAR_POSITIONAL)

        self.assertTrue(repr(p).startswith("<Parameter"))
        self.assertTrue('"a=42"' in repr(p))

    def test_signature_parameter_hashable(self):
        P = inspect.Parameter
        foo = P("foo", kind=P.POSITIONAL_ONLY)
        self.assertEqual(hash(foo), hash(P("foo", kind=P.POSITIONAL_ONLY)))
        self.assertNotEqual(
            hash(foo), hash(P("foo", kind=P.POSITIONAL_ONLY, default=42))
        )
        self.assertNotEqual(hash(foo), hash(foo.replace(kind=P.VAR_POSITIONAL)))

    def test_signature_parameter_equality(self):
        P = inspect.Parameter
        p = P("foo", default=42, kind=inspect.Parameter.KEYWORD_ONLY)

        self.assertTrue(p == p)
        self.assertFalse(p != p)
        self.assertFalse(p == 42)
        self.assertTrue(p != 42)
        self.assertTrue(p == ALWAYS_EQ)
        self.assertFalse(p != ALWAYS_EQ)

        self.assertTrue(p == P("foo", default=42, kind=inspect.Parameter.KEYWORD_ONLY))
        self.assertFalse(p != P("foo", default=42, kind=inspect.Parameter.KEYWORD_ONLY))

    def test_signature_parameter_replace(self):
        p = inspect.Parameter("foo", default=42, kind=inspect.Parameter.KEYWORD_ONLY)

        self.assertIsNot(p, p.replace())
        self.assertEqual(p, p.replace())

        p2 = p.replace(annotation=1)
        self.assertEqual(p2.annotation, 1)
        p2 = p2.replace(annotation=p2.empty)
        self.assertEqual(p, p2)

        p2 = p2.replace(name="bar")
        self.assertEqual(p2.name, "bar")
        self.assertNotEqual(p2, p)

        with self.assertRaisesRegex(ValueError, "name is a required attribute"):
            p2 = p2.replace(name=p2.empty)

        p2 = p2.replace(name="foo", default=None)
        self.assertIs(p2.default, None)
        self.assertNotEqual(p2, p)

        p2 = p2.replace(name="foo", default=p2.empty)
        self.assertIs(p2.default, p2.empty)

        p2 = p2.replace(default=42, kind=p2.POSITIONAL_OR_KEYWORD)
        self.assertEqual(p2.kind, p2.POSITIONAL_OR_KEYWORD)
        self.assertNotEqual(p2, p)

        with self.assertRaisesRegex(
            ValueError,
            "value <class 'Aspidites._vendor.contracts.inspection._empty'> "
            "is not a valid Parameter.kind",
        ):
            p2 = p2.replace(kind=p2.empty)

        p2 = p2.replace(kind=p2.KEYWORD_ONLY)
        self.assertEqual(p2, p)

    def test_signature_parameter_positional_only(self):
        with self.assertRaisesRegex(TypeError, "name must be a str"):
            inspect.Parameter(None, kind=inspect.Parameter.POSITIONAL_ONLY)

    @cpython_only
    def test_signature_parameter_implicit(self):
        with self.assertRaisesRegex(
            ValueError,
            "implicit arguments must be passed as "
            "positional or keyword arguments, "
            "not positional-only",
        ):
            inspect.Parameter(".0", kind=inspect.Parameter.POSITIONAL_ONLY)

        param = inspect.Parameter(".0", kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
        self.assertEqual(param.kind, inspect.Parameter.POSITIONAL_ONLY)
        self.assertEqual(param.name, "implicit0")

    def test_signature_parameter_immutability(self):
        p = inspect.Parameter("spam", kind=inspect.Parameter.KEYWORD_ONLY)

        with self.assertRaises(AttributeError):
            p.foo = "bar"

        with self.assertRaises(AttributeError):
            p.kind = 123


if __name__ == "__main__":
    unittest.main()



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected
from ..syntax import W, contract_expression, add_contract, add_keyword, Keyword
from Aspidites._vendor.pyparsing import (
    Dict,
    delimitedList,
    Group,
    alphanums,
    Suppress,
    Literal,
    Word,
)


class Attr(Contract):
    def __init__(self, attrs, where=None):
        # attrs: dict(str: contract)
        Contract.__init__(self, where)
        if not attrs:
            raise ValueError("Need some attributes")
        self.attrs = attrs

    def check_contract(self, context, value, silent):
        for k in self.attrs:
            if hasattr(value, k):
                self.attrs[k]._check_contract(context, getattr(value, k), silent)
            else:
                error = "Expected an attribute %r." % k
                raise ContractNotRespected(
                    contract=self, error=error, value=value, context=context
                )

    def __str__(self):
        sattrs = ";".join(["%s:%s" % (k, self.attrs[k]) for k in sorted(self.attrs)])
        return "attr(%s)" % sattrs

    def __repr__(self):
        rattrs = ",".join(["%r:%r" % (k, self.attrs[k]) for k in sorted(self.attrs)])
        return "Attr({%s})" % rattrs

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        attrs = tokens.get("attrs", None)
        # Python 3 only
        # attrs = {k: v for k, v in attrs.items()}
        attrs = dict([(k, v) for k, v in attrs.items()])
        return Attr(attrs, where=where)


attr_spec = Dict(
    delimitedList(
        Group(Word(alphanums) + Suppress(Literal(":")) + contract_expression("value")),
        delim=";",
    )
)("attrs")
attrs_spec = "(" - attr_spec - ")"


attr_contract = Keyword("attr") - attrs_spec
attr_contract.setParseAction(Attr.parse_action)

add_contract(attr_contract)
add_keyword("attr")



from __future__ import annotations

import re
import unittest
from dataclasses import dataclass

from Aspidites._vendor.apm import *


class TerseStyleTest(unittest.TestCase):
    def test_fibonacci(self):
        def fibonacci(n):
            return match(
                n, 1, 1, 2, 1, _, lambda x: fibonacci(x - 1) + fibonacci(x - 2)
            )

        self.assertEqual(1, fibonacci(1))
        self.assertEqual(1, fibonacci(2))
        self.assertEqual(2, fibonacci(3))
        self.assertEqual(3, fibonacci(4))
        self.assertEqual(5, fibonacci(5))
        self.assertEqual(8, fibonacci(6))

    def test_underscore_as_key(self):
        pet = {"type": "dog", "details": {"age": 3}}

        self.assertEqual(3, match(pet, {"details": {"age": _}}, lambda age: age))
        self.assertEqual(
            ("details", 3), match(pet, {_: {"age": _}}, lambda a, b: (a, b))
        )
        self.assertEqual(
            ("details", 3), match(pet, {_ >> "key": {"age": _}}, lambda a, b: (a, b))
        )

    def test_match_class_hierarchies(self):
        class Animal:
            pass

        class Hippo(Animal):
            pass

        class Zebra(Animal):
            pass

        class Horse(Animal):
            pass

        def what_is(x):
            return match(
                x,
                Hippo,
                "hippo",
                Zebra,
                "zebra",
                Animal,
                "some other animal",
                _,
                "not at all an animal",
            )

        self.assertEqual("zebra", what_is(Zebra()))
        self.assertEqual("hippo", what_is(Hippo()))
        self.assertEqual("some other animal", what_is(Horse()))
        self.assertEqual("some other animal", what_is(Animal()))
        self.assertEqual("not at all an animal", what_is(42))

    def test_dataclasses(self):
        @dataclass
        class Pet:
            name: str
            age: int

        pet = Pet("rover", 7)

        self.assertEqual(7, match(pet, Pet("rover", _), lambda age: age))
        self.assertEqual("rover", match(pet, Pet(_, 7), lambda name: name))
        self.assertEqual(
            ("rover", 7), match(pet, Pet(_, _), lambda name, age: (name, age))
        )

    def test_so_many_things(self):
        expectations = [
            (3, "this matches the number 3"),
            (4, "matches any integer"),
            (("k", 2), "a tuple (a, b) you can use in a function"),
            ([1, 2, 7], "any list of 3 elements that begins with [1, 2]"),
            ({"x": 3}, "any dict with a key 'x' and any value associated"),
            ({"x": 4, "y": 4}, "any dict with a key 'x' and any value associated"),
            ("xyz", "anything else"),
        ]

        for val, expected in expectations:
            result = match(
                val,
                3,
                "this matches the number 3",
                int,
                "matches any integer",
                (str, int),
                lambda a, b: "a tuple (a, b) you can use in a function",
                [1, 2, _],
                "any list of 3 elements that begins with [1, 2]",
                {"x": _},
                "any dict with a key 'x' and any value associated",
                _,
                "anything else",
            )
            self.assertEqual(expected, result)

    def test_regex(self):
        def what_is(pet):
            return match(
                pet,
                re.compile(r"(\w+)-(\w+)-cat$"),
                lambda name, my: "cat " + name,
                re.compile(r"(\w+)-(\w+)-dog$"),
                lambda name, my: "dog " + name,
                _,
                "something else",
            )

        self.assertEqual("dog fuffy", what_is("fuffy-my-dog"))
        self.assertEqual("dog puffy", what_is("puffy-her-dog"))
        self.assertEqual("cat carla", what_is("carla-your-cat"))
        self.assertEqual("something else", what_is("roger-my-hamster"))

    def test_lambda_with_too_little_parameters(self):
        self.assertEqual("anything", match(1, _, lambda: "anything"))

    def test_lambda_with_too_many_parameters(self):
        self.assertEqual(
            "anything(1, None, None)",
            match(1, _, lambda a, b, c: f"anything({a}, {b}, {c})"),
        )

    def test_default_value(self):
        self.assertEqual("fallback", match(1, 2, lambda: "something", "fallback"))

    def test_default_action(self):
        self.assertEqual(
            "fallback", match(1, 2, lambda: "something", lambda: "fallback")
        )



from . import syntax_fail, good, fail, semantic_fail

# basic types

good("int", 1)
fail("int", None)
fail("int", 2.0)
good("float", 1.1)
fail("float", None)
fail("float", 2)

good("number", 1)
good("number", 1.0)
fail("number", [1])

good("bool", False)
good("bool", True)
fail("bool", 1)
fail("bool", 0)

# type contract

good("None", None)
good("NoneType", None)
fail("None", 1)
fail("NoneType", 1)

syntax_fail("type")
syntax_fail("type()")
good("type(x)", 1)
semantic_fail("type(X)", 1)



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval


def test_In():
    assert restricted_eval("1 in [1, 2, 3]") is True


def test_NotIn():
    assert restricted_eval("4 not in [1, 2, 3]") is True



from . import good, fail, syntax_fail


class A(object):
    a = 1
    b = 2


tc_a = A()


class B(object):
    a = 2


tc_b = B()


syntax_fail("attr")  # need at least some attribute

good("attr(a:*)", tc_a)
good("attr(a:int)", tc_a)
good("attr(b:int)", tc_a)
good("attr(b:>1)", tc_a)
good("attr(b:int,>1)", tc_a)
fail("attr(b:int,<=1)", tc_a)

good("attr(a:*)", tc_b)
fail("attr(b:*)", tc_b)

good("attr(a:int;b:int)", tc_a)
good("attr(a:int;b:int)", tc_a)
good("attr(a:int;b:int,>1)", tc_a)
fail("attr(a:int;b:int,<=1)", tc_a)



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected
from ..syntax import add_contract, W, contract_expression, O, S, add_keyword, Keyword
from Aspidites._vendor.pyrsistent import PVector


class List(Contract):
    def __init__(self, length_contract=None, elements_contract=None, where=None):
        Contract.__init__(self, where)
        self.length_contract = length_contract
        self.elements_contract = elements_contract

    def check_contract(self, context, value, silent):
        if not isinstance(value, (list, PVector)):
            error = "Expected a list, got %r." % value.__class__.__name__
            raise ContractNotRespected(self, error, value, context)

        if self.length_contract is not None:
            self.length_contract._check_contract(context, len(value), silent)

        if self.elements_contract is not None:
            for element in value:
                self.elements_contract._check_contract(context, element, silent)

    def __str__(self):
        s = "list"
        if self.length_contract is not None:
            s += "[%s]" % self.length_contract
        if self.elements_contract is not None:
            s += "(%s)" % self.elements_contract
        return s

    def __repr__(self):
        s = "List(%r,%r)" % (self.length_contract, self.elements_contract)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        length_contract = tokens.get("length_contract", None)
        elements_contract = tokens.get("elements_contract", None)
        return List(length_contract, elements_contract, where=where)


list_contract = (
    Keyword("list")
    - O(S("[") - contract_expression("length_contract") - S("]"))
    + O(S("(") - contract_expression("elements_contract") - S(")"))
)
list_contract.setParseAction(List.parse_action)

list_contract.setName("List contract")
add_keyword("list")
add_contract(list_contract)



from . import syntax_fail, good, fail, semantic_fail

# only single letters
syntax_fail("NN")
syntax_fail("xx")

# big letters can only bind to numbers
good("N,N>0", 1)
fail("N,N>0", 0)
semantic_fail("N", [])

# lower case can bind to anything
good("x", 1)
good("x", [])



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=True, initializedcheck=False
# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .._vendor.pyparsing import (
    IndentedBlock,
    Combine,
    Empty,
    FollowedBy,
    Forward,
    ZeroOrMore,
    OneOrMore,
    MatchFirst,
    Optional,
    ParseElementEnhance,
    ParseException,
    Regex,
    Char,
    Word,
    alphanums,
    alphas,
    col,
    conditionAsParseAction,
    delimitedList,
    infixNotation,
    matchOnlyAtCol,
    nums,
    opAssoc,
    quotedString,
    unicodeString,
)

from .._vendor.contracts.syntax import contract_expression
from .._vendor.contracts.library.suggester import create_suggester
from ..parser.convert import *

list_str = Forward().setName("list literal")
list_str_evolver = Forward().setName("list evolver literal")
list_index = Forward().setName("list index operation")
list_set = Forward().setName("list set operation")
list_count = Forward().setName("list count operation")
list_append = Forward().setName("list append operation")
list_remove = Forward().setName("list remove operation")
dict_items = Forward().setName("dictionary items operation")
dict_keys = Forward().setName("dictionary keys operation")
dict_vals = Forward().setName("dictionary values operation")
dict_discard = Forward().setName("dictionary discard operation")
dict_update = Forward().setName("dictionary update operation")
dict_copy = Forward().setName("dictionary copy operation")
dict_remove = Forward().setName("dictionary remove operation")
set_discard = Forward().setName("set discard operation")
set_update = Forward().setName("set update operation")
set_copy = Forward().setName("set copy operation")
set_remove = Forward().setName("set remove operation")
set_str = Forward().setName("set literal")
set_str_evolver = Forward().setName("set evolver literal")
dict_str = Forward().setName("dictionary literal")
dict_str_evolver = Forward().setName("dictionary evolver literal")
# tuple_str = Forward()
slice_str = Forward().setName("list slice operation")
simple_assign = Forward().setName("simple assignment")
slice_assign = Forward().setName("slice assignment")
func_call = Forward().setName("function call")
suite = Forward().setName("indented block")
break_if = Forward().setName("conditional break")
cont_if = Forward().setName("conditional continue")
if_slice = Forward().setName("conditional slice")
if_simple = Forward().setName("conditional assignment")
if_func_call = Forward().setName("conditional function call")
if_return = Forward().setName("conditional return")
rvalue = Forward().setName("return value")
stmt = Forward().setName("statement")

quoted_str = quotedString().setParseAction(lambda s, l, t: t[0])
quoted_str.setName("string literal")
integer = Word(nums).setParseAction(cvt_int)
integer.setName("integer numeric literal")
real = Combine(Optional(Word(nums)) + "." + Word(nums))
real.setName("floating point numeric literal")
complex_ = Combine(real | integer + "+" + real | integer + "j")
complex_.setName("complex numeric literal")
identifier = Word(alphas + "_", alphanums + "_") + Optional(persist)
identifier.setName("identifier")
operand = nullit | complex_ | real | bool_literal | integer | identifier | underscore
operand.setName("operand")
arith_expr = Combine(
    infixNotation(
        operand,
        [
            (factop, 1, opAssoc.LEFT),
            (exponp, 2, opAssoc.RIGHT),
            (
                signop,
                1,
                opAssoc.RIGHT,
            ),  # Changing this to LEFT causes the entire parser to break
            (multop, 2, opAssoc.LEFT),
            (plusop, 2, opAssoc.LEFT),
            (bitwiseop, 2, opAssoc.LEFT),
        ],
        lpar=lit_lparen,
        rpar=lit_rparen,
    )
).setParseAction(cvt_arith_expr)
arith_expr.setName("arithmetic expression")
comp_expr = infixNotation(
    arith_expr,
    [
        (comparisonop, 2, opAssoc.LEFT),
    ],
).setParseAction(lambda s, l, t: "".join(t[0]))
comp_expr.setName("comparison expression")
# TODO dict trigrams cause RecursionError:
#  maximum recursion depth exceeded while getting the str of an object
list_item = (  # Precedence important!!!
    slice_str
    | comp_expr  # Expressions
    | arith_expr
    | identifier
    | integer  # Literals
    | complex_
    | real
    | bool_literal
    | nullit
    | quoted_str
    | list_str_evolver
    | set_str_evolver
    | dict_str_evolver
    | list_str
    | set_str
    | dict_str
).setName("list item")
collection_trigrams = (
    list_set
    | list_index
    | list_count
    | dict_discard
    | set_discard
    | list_append
    | dict_copy
    | set_copy
    | list_remove
    | dict_remove
    | set_remove
).setName("collections trigram")
list_index <<= (
    (identifier | list_str | slice_str)
    + index_items.setParseAction(replaceWith(".index"))
    + list_item
).setParseAction(cvt_list_index)

list_count <<= (
    (identifier | list_str | slice_str)
    + count_vals.setParseAction(replaceWith(".count"))
    + list_item
).setParseAction(cvt_list_index)

list_set <<= (
    identifier
    + set_add.setParseAction(replaceWith(".set"))
    + identifier
    + ","
    + identifier
).setParseAction(cvt_list_index)

list_append <<= (
    (identifier | list_str | slice_str)
    + append.setParseAction(replaceWith(".append"))
    + list_item
).setParseAction(cvt_list_index)

list_remove <<= (
    (identifier | list_str | slice_str)
    + remove.setParseAction(replaceWith(".remove"))
    + list_item
).setParseAction(cvt_list_index)

# dict_items <<= ((dict_str) + index_items.setParseAction(replaceWith('.items()')))
# dict_keys <<= ((dict_str) + keys.setParseAction(replaceWith('.keys()')))
# dict_vals <<= ((dict_str) + count_vals.setParseAction(replaceWith('.values()')))
dict_discard <<= (
    (dict_str) + discard.setParseAction(replaceWith(".discard")) + list_item
).setParseAction(cvt_list_index)
dict_update <<= (
    (identifier | dict_str) + update.setParseAction(replaceWith(".update")) + list_item
).setParseAction(cvt_list_index)

dict_copy <<= (identifier | dict_str) + copy.setParseAction(replaceWith(".copy()"))

dict_remove <<= (
    (identifier | dict_str) + remove.setParseAction(replaceWith(".remove")) + list_item
).setParseAction(cvt_list_index)

# set_discard <<= (
#         (identifier | set_str) + discard.setParseAction(
#     replaceWith('.discard')) + list_item).setParseAction(
#     cvt_list_index)
# set_update <<= (
#         (identifier | set_str) + append.setParseAction(
#     replaceWith('.append')) + list_item).setParseAction(
#     cvt_list_index)

set_copy <<= (identifier | set_str) + copy.setParseAction(replaceWith(".copy()"))

set_remove <<= (
    (identifier | set_str) + remove.setParseAction(replaceWith(".remove")) + list_item
).setParseAction(cvt_list_index)

lit_ellipse = Literal("...").setParseAction(replaceWith("..."))
assignable = collection_trigrams | func_call | list_item | lit_ellipse

_contract_expression = contract_expression.copy()
_contract_expression.setParseAction(
    lambda s, l, tks: f"'{''.join((str(t) for t in tks))}'"
)
contract_define = identifier + imposes + _contract_expression
contract_define.setParseAction(cvt_contract_define)
contract_respect = respects + _contract_expression
contract_assign = identifier + assign_eq + assignable + contract_respect
contract_assign.setParseAction(cvt_contract_assign)

# tuple_str <<= (lparen + Optional(delimitedList(list_item)) + Optional(comma) + rparen).setParseAction(cvt_tuple)
noclosure.setParseAction(replaceWith(".evolver()"))
list_str <<= (
    lbrack + Optional(delimitedList(list_item) + Optional(comma)) + rbrack
).setParseAction(cvt_list)
list_str_evolver <<= (
    (
        lbrack + Optional(delimitedList(list_item) + Optional(comma)) + rbrack
    ).setParseAction(cvt_list)
    + noclosure
).setParseAction(lambda s, l, t: "".join(t))
set_str <<= (
    lbrace + Optional(delimitedList(list_item) + Optional(comma)) + rbrace
).setParseAction(cvt_set)
set_str_evolver <<= (
    (
        lbrace + Optional(delimitedList(list_item) + Optional(comma)) + rbrace
    ).setParseAction(cvt_set)
    + noclosure
).setParseAction(lambda s, l, t: "".join(t))
dict_entry = Group(list_item + colon + list_item)
dict_str <<= (
    lbrace + Optional(delimitedList(dict_entry) + Optional(comma)) + rbrace
).setParseAction(cvt_dict)
dict_str_evolver <<= (
    (
        lbrace + Optional(delimitedList(dict_entry) + Optional(comma)) + rbrace
    ).setParseAction(cvt_dict)
    + noclosure
).setParseAction(lambda s, l, t: "".join(t))
slice_str <<= (
    identifier
    + lit_lbrack
    + Optional(integer | identifier)
    + Optional(lit_colon)
    + Optional(integer | identifier)
    + Optional(lit_colon)
    + Optional(integer | identifier)
    + lit_rbrack
)
slice_str.setParseAction(lambda s, l, t: "".join(str(i) for i in t))
def_args = Optional(delimitedList(contract_assign, delim=";")).setParseAction(
    lambda s, l, t: sep.join(t)
)
def_args = Group(lit_lparen + def_args + args_end).setParseAction(
    lambda s, l, t: "".join(*t)
)

bool_pragmas = Combine(
    pragma
    + oneOf(" ".join(available_bool_pragmas))
    + lit_lparen
    + oneOf("True False")
    + lit_rparen
).setParseAction(cvt_pragma)

pragmas = Combine(pragma + oneOf(" ".join(available_pragmas))).setParseAction(
    cvt_pragma
)

func_decl = Group(
    Optional(OneOrMore(bool_pragmas) | OneOrMore(pragmas))
    + private_def_decl
    + identifier
    + def_args
    + _contract_expression
).setParseAction(lambda s, l, t: "\n@__contract()\n" + "".join(*t) + lit_colon)
func_decl.setName("function declaration")
comment_line = (
    Combine(Regex(r"`(?:[^`\n\r\\]|(?:``)|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*") + "`")
    .setParseAction(lambda s, l, t: t[0])
    .setParseAction(cvt_comment_line)
)

return_value = return_none + rvalue
yield_value = yield_none + rvalue
ret_stmt = Group(return_value).setParseAction(lambda s, l, t: "".join(*t))
yield_stmt = Group(yield_value).setParseAction(lambda s, l, t: "".join(*t))

func_def = Group(func_decl + suite).setParseAction(
    lambda s, l, t: nl_indent.join(t[0]) + "\n"
)
lambda_def = Combine(Group(lit_lparen + arith_expr | comp_expr + lit_rparen))
func_call <<= Group(
    identifier + lit_lparen + Optional(delimitedList(rvalue)) + lit_rparen
).setParseAction(
    cvt_func_call
)  # if len(t[0]) != 3 else t[0][0] + '()')
clos_call = Group(
    identifier + lit_lparen + Optional(delimitedList(rvalue)) + lit_rparen
).setParseAction(cvt_clos_call) + Suppress(noclosure)
context_suite = Forward()
context_def = Group(context_stmt + func_call).setParseAction(
    lambda s, l, t: "".join(*t) + lit_colon
)

context_decl = Group(context_def + context_suite).setParseAction(
    lambda s, l, t: "".join(*t)
)

# TODO (!): trigram only binds a single letter variable identifier
case_stmt = Group(assignable + colon + func_call).setParseAction(
    lambda s, l, t: sep.join(t[0])
)
match_suite = Group(IndentedBlock(OneOrMore(case_stmt))).setParseAction(
    lambda s, l, t: (sep.join(t.asList()[0]))
)
match_decl = Group(match_none + identifier).setParseAction(
    lambda s, l, t: t[0][1] + "=" + t[0][0] + lit_lparen + t[0][1]
)
match_def = Group(match_decl + match_suite).setParseAction(
    lambda s, l, t: sep.join(t[0]) + lit_rparen
)

loop_suite = Forward()

ident_loop_decl = Group(
    identifier + Optional(lit_comma + identifier) + for_none + identifier
).setParseAction(cvt_for_loop_decl)
ident_loop_def = Group(ident_loop_decl + loop_suite).setParseAction(
    lambda s, l, t: (nl_indent + indent).join(t[0])
)

list_loop_decl = Group(
    identifier + Optional(lit_comma + identifier) + for_none + list_str
).setParseAction(cvt_for_loop_decl)
list_loop_def = Group(list_loop_decl + loop_suite).setParseAction(
    lambda s, l, t: (nl_indent + indent).join(t[0])
)

set_loop_decl = Group(
    identifier + Optional(lit_comma + identifier) + for_none + set_str
).setParseAction(cvt_for_loop_decl)
set_loop_def = Group(set_loop_decl + loop_suite).setParseAction(
    lambda s, l, t: (nl_indent + indent).join(t[0])
)

# tuple_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + tuple_str).setParseAction(
#     cvt_for_loop_decl)
# tuple_loop_def = Group(tuple_loop_decl + loop_suite).setParseAction(lambda s, l, t: (nl_indent + indent).join(t[0]))

dict_loop_decl = Group(
    identifier + Optional(lit_comma + identifier) + for_none + dict_str
).setParseAction(cvt_for_loop_decl)
dict_loop_def = Group(dict_loop_decl + loop_suite).setParseAction(
    lambda s, l, t: (nl_indent + indent).join(t[0])
)

string_loop_decl = Group(
    identifier + Optional(lit_comma + identifier) + for_none + quoted_str
).setParseAction(cvt_for_loop_decl)
string_loop_def = Group(string_loop_decl + loop_suite).setParseAction(
    lambda s, l, t: (nl_indent + indent).join(t[0])
)

func_loop_decl = Group(
    identifier + Optional(lit_comma + identifier) + for_none + func_call
).setParseAction(cvt_for_loop_decl)
func_loop_def = Group(func_loop_decl + loop_suite).setParseAction(
    lambda s, l, t: (nl_indent + indent).join(t[0])
)
break_if <<= (break_stmt + rvalue).setParseAction(
    lambda s, l, t: "if " + t[1] + lit_colon + t[0]
)
cont_if <<= (cont_stmt + rvalue).setParseAction(
    lambda s, l, t: "if " + t[1] + lit_colon + t[0]
)
if_slice <<= (rvalue + if_cond + slice_assign).setParseAction(
    lambda s, l, t: t[1] + t[0] + lit_colon + t[2]
)
if_simple <<= (rvalue + if_cond + simple_assign).setParseAction(
    lambda s, l, t: t[1] + t[0] + lit_colon + t[2]
)
if_func_call <<= (rvalue + if_cond + func_call).setParseAction(
    lambda s, l, t: t[1] + t[0] + lit_colon + t[2]
)
if_return <<= (rvalue + if_cond + return_value | return_none).setParseAction(
    lambda s, l, t: t[1] + t[0] + lit_colon + t[2] + t[3]
)

loop_suite <<= IndentedBlock(
    OneOrMore(
        pass_stmt
        | yield_stmt
        | cont_if
        | cont_stmt
        | break_if
        | break_stmt
        | if_slice
        | if_simple
        | if_func_call
        | if_return
        | func_call
        | simple_assign
        | slice_assign
    )
)
# TODO context managers get eaten by the preceding code blocks
context_suite <<= IndentedBlock(
    OneOrMore(contract_assign | match_def | func_call)
).setParseAction(lambda s, l, t: (nl_indent.join(t.asList())))

suite <<= IndentedBlock(
    OneOrMore(
        pass_stmt
        | ret_stmt
        | yield_stmt
        | ident_loop_def
        | string_loop_def
        | list_loop_def
        | set_loop_def
        # | tuple_loop_def
        | dict_loop_def
        | func_loop_def
        | match_def
        | if_slice
        | if_simple
        | if_func_call
        | if_return
        | slice_assign
        | contract_assign
    )
).setParseAction(lambda s, l, t: (nl_indent.join(t.asList())))
rvalue <<= collection_trigrams | clos_call | func_call | list_item | lambda_def
simple_assign << Group(identifier + assign_eq + rvalue).setParseAction(
    lambda s, l, t: " ".join(t[0])
)
slice_assign << Group(slice_str + assign_eq + rvalue).setParseAction(
    lambda s, l, t: " ".join(t[0])
)
stmt <<= func_def | contract_define | simple_assign
module_body = OneOrMore(stmt) + Optional(
    struct_main
    + OneOrMore(stmt).setParseAction(lambda s, l, t: indent + nl_indent.join(t))
)
module_body.ignore(comment_line)
module_body.setName("module")


def parse_module(module):
    return module_body.parseString(module, parseAll=True)


def parse_statement(s):
    return stmt.parseString(s)


# if __name__ == "__main__":
#     print(parse_module(open('../examples/math.wom').read()))



# line 1
def wrap(foo=None):
    def wrapper(func):
        return func

    return wrapper


# line 7
def replace(func):
    def insteadfunc():
        print("hello")

    return insteadfunc


# line 13
@wrap()
@wrap(wrap)
def wrapped():
    pass


# line 19
@replace
def gone():
    pass


# line 24
oll = lambda m: m

# line 27
tll = lambda g: g and g and g

# line 32
tlli = lambda d: d and d

# line 36
def onelinefunc():
    pass


# line 39
def manyargs(arg1, arg2, arg3, arg4):
    pass


# line 43
def twolinefunc(m):
    return m and m


# line 47
a = [None, lambda x: x, None]

# line 52
def setfunc(func):
    globals()["anonymous"] = func


setfunc(lambda x, y: x * y)

# line 57
def with_comment():  # hello
    world


# line 61
multiline_sig = [
    lambda x, y: x + y,
    None,
]

# line 68
def func69():
    class cls70:
        def func71():
            pass

    return cls70


extra74 = 74

# line 76
def func77():
    pass


(extra78, stuff78) = "xy"
extra79 = "stop"

# line 81
class cls82:
    def func83():
        pass


(extra84, stuff84) = "xy"
extra85 = "stop"

# line 87
def func88():
    # comment
    return 90


# line 92
def f():
    class X:
        def g():
            "doc"
            return 42

    return X


method_in_dynamic_class = f().g

# line 101
def keyworded(*arg1, arg2=1):
    pass


# line 105
def annotated(arg1: list):
    pass


# line 109
def keyword_only_arg(*, arg):
    pass


@wrap(lambda: None)
def func114():
    return 115


class ClassWithMethod:
    def method(self):
        pass


from functools import wraps


def decorator(func):
    @wraps(func)
    def fake():
        return 42

    return fake


# line 129
@decorator
def real():
    return 20


# line 134
class cls135:
    def func136():
        def func137():
            never_reached1
            never_reached2


# line 141
class cls142:
    a = """
class cls149:
    ...
"""


# line 148
class cls149:
    def func151(self):
        pass


"""
class cls160:
    pass
"""

# line 159
class cls160:
    def func162(self):
        pass


# line 165
class cls166:
    a = """
    class cls175:
        ...
    """


# line 172
class cls173:
    class cls175:
        pass


# line 178
class cls179:
    pass


# line 182
class cls183:
    class cls185:
        def func186(self):
            pass


def class_decorator(cls):
    return cls


# line 193
@class_decorator
@class_decorator
class cls196:
    @class_decorator
    @class_decorator
    class cls200:
        pass


class cls203:
    class cls204:
        class cls205:
            pass

    class cls207:
        class cls205:
            pass


# line 211
def func212():
    class cls213:
        pass

    return cls213


# line 217
class cls213:
    def func219(self):
        class cls220:
            pass

        return cls220


# line 224
async def func225():
    class cls226:
        pass

    return cls226


# line 230
class cls226:
    async def func232(self):
        class cls233:
            pass

        return cls233


if True:

    class cls238:
        class cls239:
            """if clause cls239"""


else:

    class cls238:
        class cls239:
            """else clause 239"""

            pass


# line 247
def positional_only_arg(a, /):
    pass


# line 251
def all_markers(a, b, /, c, d, *, e, f):
    pass


# line 255
def all_markers_with_args_and_kwargs(a, b, /, c, d, *args, e, f, **kwargs):
    pass


# line 259
def all_markers_with_defaults(a, b=1, /, c=2, d=3, *, e=4, f=5):
    pass



__version__ = "1.15.7"
__title__ = "Aspidites"
__author__ = "Ross J. Duff"
__license__ = "GPL"
__description__ = f"""
    {__title__} v{__version__} is the reference Woma programming language compiler.\n
    Copyright (C) 2021  {__author__}\n
    This program comes with ABSOLUTELY NO WARRANTY;\n
    This is free software, and you are welcome to redistribute it\n
    under the conditions of the {__license__}v3 found in the LICENSE file.
    """
__mimetype__ = "text/woma"
__all__ = [
    "SafeMod",
    "SafeDiv",
    "SafeExp",
    "SafeFloorDiv",
    "SafeFactorial",
    "SafeUnarySub",
    "SafeUnaryAdd",
    "Undefined",
    "Maybe",
    "Surely",
    "Warn",
    "parse_module",
    "parse_statement",
    "bordered",
]

from Aspidites.api import *
from Aspidites.parser import *
from Aspidites.monads import *
from Aspidites.math import *

Maybe.__doc__ = Maybe.__doc__



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
import collections.abc as collections


def ist(C):
    def f(x):
        f.__name__ = "isinstance_of_%s" % C.__name__
        if not isinstance(x, C):
            raise ValueError("Value is not an instance of %s." % C.__name__)

    return f


def m_new_contract(name, f):
    from Aspidites._vendor.contracts.library.extensions import CheckCallable
    from Aspidites._vendor.contracts.library.extensions import Extension

    Extension.registrar[name] = CheckCallable(f)


m_new_contract("Container", ist(collections.Container))
m_new_contract("Iterable", ist(collections.Iterable))

m_new_contract("Hashable", ist(collections.Hashable))
m_new_contract("Generator", ist(collections.Generator))


m_new_contract("Iterator", ist(collections.Iterator))
m_new_contract("Sized", ist(collections.Sized))
m_new_contract("Callable", ist(collections.Callable))
m_new_contract("Sequence", ist(collections.Sequence))
m_new_contract("Set", ist(collections.Set))
m_new_contract("MutableSequence", ist(collections.MutableSequence))
m_new_contract("MutableSet", ist(collections.MutableSet))
m_new_contract("Mapping", ist(collections.Mapping))
m_new_contract("MutableMapping", ist(collections.MutableMapping))
# new_contract('MappingView', ist(collections.MappingView))
# new_contract('ItemsView', ist(collections.ItemsView))
# new_contract('ValuesView', ist(collections.ValuesView))


# Not a lambda to have better messages
def is_None(x):
    return x is None


m_new_contract("procedure", is_None)
m_new_contract("coroutine", ist(collections.Generator))
m_new_contract("None", is_None)
m_new_contract("NoneType", is_None)



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ...._vendor._compat import basestring

from ..interface import Contract, ContractNotRespected, RValue, describe_value
from ..syntax import W, oneOf, FollowedBy, NotAny


class BindVariable(Contract):
    def __init__(self, variable, allowed_types, where=None):
        assert isinstance(variable, basestring) and len(variable) == 1
        assert allowed_types, "%r" % allowed_types
        Contract.__init__(self, where)
        self.variable = variable
        self.allowed_types = allowed_types

    def check_contract(self, context, value, silent):
        if self.variable in context:
            expected = context[self.variable]
            if not (expected == value):
                error = (
                    "Expected value for %r was: %s\n        instead I received: %s"
                    % (self.variable, describe_value(expected), describe_value(value))
                )
                raise ContractNotRespected(
                    contract=self, error=error, value=value, context=context
                )

        else:
            # bound variable
            if not isinstance(value, self.allowed_types):
                error = "Variable %r can only bind to %r, not %r." % (
                    self.variable,
                    self.allowed_types,
                    value.__class__.__name__,
                )
                raise ContractNotRespected(self, error, value, context)

            context[self.variable] = value

    def __str__(self):
        return self.variable

    def __repr__(self):
        # XXX: invalid if tuple
        return "BindVariable(%r,%s)" % (self.variable, self.allowed_types.__name__)

    @staticmethod
    def parse_action(allowed_types):
        def parse(s, loc, tokens):
            where = W(s, loc)
            variable = tokens[0]
            assert len(variable) == 1, "Wrong syntax, matched %r as variable in %r." % (
                variable,
                s,
            )
            # print ('Matched %r as variable in %r.' % (variable, s))
            return BindVariable(variable, allowed_types, where=where)

        return parse


class VariableRef(RValue):
    def __init__(self, variable, where=None):
        assert isinstance(variable, basestring)
        self.where = where
        self.variable = variable

    def eval(self, context):  # @ReservedAssignment
        var = self.variable
        if not var in context:
            raise ValueError("Unknown variable %r." % var)
        return context[var]

    def __repr__(self):
        return "VariableRef(%r)" % self.variable

    def __str__(self):
        return "%s" % self.variable

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        return VariableRef(tokens[0], where=where)


alphabetu = "A B C D E F G H I J K L M N O P Q R S T U W V X Y Z "
alphabetl = "a b c d e f g h i j k l m n o p q r s t u w v x y z "

# Special case: allow an expression like AxBxC
nofollow = "a b c d e f g h i j k l m n o p q r s t u w v   y z"
# also do not commit if part of word (SEn, a_2)
nofollow += " A B C D E F G H I J K L M N O P Q R S T U W V X Y Z "
nofollow += " 0 1 2 3 4 5 6 7 8 9 _"
# but recall 'axis_angle'
int_variables = oneOf(alphabetu.split()) + FollowedBy(NotAny(oneOf(nofollow.split())))
misc_variables = oneOf(alphabetl.split()) + FollowedBy(
    NotAny(oneOf(nofollow.split() + ["x"]))
)
int_variables_ref = int_variables.copy().setParseAction(VariableRef.parse_action)
misc_variables_ref = misc_variables.copy().setParseAction(VariableRef.parse_action)

# int_variables = oneOf(alphabetu.split()) + FollowedBy(White() ^ 'x')

# These must be followed by whitespace; punctuation
# misc_variables = oneOf(alphabet.lower()) + FollowedBy(White())

nofollow = "a b c d e f g h i j k l m n o p q r s t u w v   y z "
nofollow += " * - + / "
nofollow += " A B C D E F G H I J K L M N O P Q R S T U W V X Y Z "
nofollow += " 0 1 2 3 4 5 6 7 8 9 _"
int_variables2 = oneOf(alphabetu.split()) + FollowedBy(NotAny(oneOf(nofollow.split())))
misc_variables2 = oneOf(alphabetl.split()) + FollowedBy(
    NotAny(oneOf(nofollow.split() + ["x"]))
)
int_variables_contract = int_variables2.setParseAction(BindVariable.parse_action(int))
misc_variables_contract = misc_variables2.setParseAction(
    BindVariable.parse_action(object)
)



from . import syntax_fail, good, fail

good("dict", {})
fail("dict", 1)
fail("dict", [])
syntax_fail("dict[]")
syntax_fail("dict[]()")
syntax_fail("dict()")
good("dict[1]", {1: 2})
good("dict[N],N<2", {1: 2})
fail("dict[N],N<2", {1: 2, 3: 4})
good("dict(int:int)", {1: 2})
fail("dict(int:int)", {"a": 2})
good("dict(*:int)", {1: 2})
good("dict(*:int)", {"a": 2})

# dictionary of string -> tuple, with tuple of two elements with different type
good("dict(str:tuple(type(x),type(y))),x!=y", {"a": (2, 1.1)})
fail("dict(str:tuple(type(x),type(y))),x!=y", {"a": (2, 1)})



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected
from ..syntax import (
    add_contract,
    W,
    contract_expression,
    O,
    S,
    ZeroOrMore,
    Group,
    add_keyword,
    Keyword,
)
from .compositions import or_contract


class Tuple(Contract):
    def __init__(self, length=None, elements=None, where=None):
        Contract.__init__(self, where)
        self.length = length
        self.elements = elements
        assert elements is None or isinstance(elements, list)
        if elements:
            for e in elements:
                assert isinstance(e, Contract)

    def check_contract(self, context, value, silent):
        if not isinstance(value, tuple):
            error = "Expected a tuple, got %r." % value.__class__.__name__
            raise ContractNotRespected(
                contract=self, error=error, value=value, context=context
            )

        if self.length is not None:
            self.length._check_contract(context, len(value), silent)

        if self.elements is not None:
            if len(value) != len(self.elements):
                error = "Expected a tuple of length " "%s, got %r of length %s." % (
                    len(self.elements),
                    value,
                    len(value),
                )
                raise ContractNotRespected(
                    contract=self, error=error, value=value, context=context
                )

            for i in range(len(value)):
                self.elements[i]._check_contract(context, value[i], silent)

    def __repr__(self):
        return "Tuple(%r,%r)" % (self.length, self.elements)

    def __str__(self):
        s = "tuple"
        if self.length is not None:
            s += "[%s]" % self.length

        def rep(x):
            from .compositions import And

            if isinstance(x, And):
                return "(%s)" % x
            else:
                return "%s" % x

        if self.elements is not None:
            s += "(%s)" % ",".join(rep(x) for x in self.elements)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        length = tokens.get("length", [None])[0]
        #        elements = tokens.get('elements', [None])[0]

        if "elements" in tokens:
            elements = list(tokens["elements"])
        else:
            elements = None
        assert elements is None or length is None
        assert length is None or isinstance(length, Contract), "Wrong type %r" % length

        if elements:
            for e in elements:
                assert isinstance(e, Contract), "Wrong type %s (%r)" % (type(e), e)
        return Tuple(length, elements, where=where)


# if you use contract instead of simple_contract, it will be matched as And


inside = (S("(") - contract_expression - S(")")) | or_contract
inside.setName("Any contract for tuple elements (use parenthesis for AND)")

elements = Group(S("(") - inside - ZeroOrMore(S(",") - inside) - S(")"))("elements")
elements.setName("Tuple elements contract.")


length = Group(S("[") - contract_expression - S("]"))("length")
length.setName("Tuple length contract.")

tuple_contract = Keyword("tuple") - O(length | elements)
tuple_contract.setName("tuple contract")

add_contract(tuple_contract.setParseAction(Tuple.parse_action))
add_keyword("tuple")



from . import good, fail
import collections

# sequences
for obj in (
    [],
    tuple(),
    set(),
    frozenset(),
    collections.deque(),
    {},
    "ciao",
):
    good("collection[*]", obj)
    good("collection[*]", obj)
    good("collection[*](*)", obj)
    good("collection[N]", obj)
    fail("collection[N],N>0", [])

for obj in (
    [1.0],
    (1.0,),
    {1.0},
    frozenset((1.0,)),
    collections.deque((1.0,)),
    {1.0: 1},
):
    good("collection[*](float)", obj)
    good("collection[=1]", obj)
    fail("collection[=2]", obj)
    good("collection[1]", obj)  # shortcut

    good("collection[N],N>0", obj)
    good("collection[N],N=1", obj)
    good("collection[N],N>0,N<2", obj)

    fail("collection[*](int)", obj)



from typing import TypeVar
from collections import namedtuple
from Aspidites._vendor._compat import IS_CPYTHON
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor.RestrictedPython.transformer import RestrictingNodeTransformer

import ast
import warnings

T = TypeVar("T")
CompileResult: T = namedtuple("CompileResult", "code, errors, warnings, used_names")
syntax_error_template = "Line {lineno}: {type}: {msg} at statement: {statement!r}"

NOT_CPYTHON_WARNING = (
    "RestrictedPython is only supported on CPython: use on other Python "
    "implementations may create security issues."
)


def _compile_restricted_mode(
    source,
    filename="<string>",
    mode="exec",
    flags=0,
    dont_inherit=False,
    policy=RestrictingNodeTransformer,
):

    if not IS_CPYTHON:
        warnings.warn_explicit(
            NOT_CPYTHON_WARNING, RuntimeWarning, "RestrictedPython", 0
        )

    byte_code = None
    collected_errors = []
    collected_warnings = []
    used_names = {}
    if policy is None:
        # Unrestricted Source Checks
        byte_code = compile(
            source, filename, mode=mode, flags=flags, dont_inherit=dont_inherit
        )
    elif issubclass(policy, RestrictingNodeTransformer):
        c_ast = None
        allowed_source_types = [str, ast.Module]
        if IS_PY2:  # pragma: no cover
            allowed_source_types.append(
                unicode
            )  # NOQA: F821,E501  # PY2 only statement, in Python 2 only module
        if not issubclass(type(source), tuple(allowed_source_types)):
            raise TypeError(
                "Not allowed source type: " '"{0.__class__.__name__}".'.format(source)
            )
        c_ast = None
        # workaround for pypy issue https://bitbucket.org/pypy/pypy/issues/2552
        if isinstance(source, ast.Module):
            c_ast = source
        else:
            try:
                c_ast = ast.parse(source, filename, mode)
            except (TypeError, ValueError) as e:
                collected_errors.append(str(e))
            except SyntaxError as v:
                collected_errors.append(
                    syntax_error_template.format(
                        lineno=v.lineno,
                        type=v.__class__.__name__,
                        msg=v.msg,
                        statement=v.text.strip() if v.text else None,
                    )
                )
        if c_ast:
            policy_instance = policy(collected_errors, collected_warnings, used_names)
            policy_instance.visit(c_ast)
            if not collected_errors:
                byte_code = compile(
                    c_ast,
                    filename,
                    mode=mode  # ,
                    # flags=flags,
                    # dont_inherit=dont_inherit
                )
    else:
        raise TypeError("Unallowed policy provided for RestrictedPython")
    return CompileResult(
        byte_code, tuple(collected_errors), collected_warnings, used_names
    )


def compile_restricted_exec(
    source,
    filename="<string>",
    flags=0,
    dont_inherit=False,
    policy=RestrictingNodeTransformer,
):
    """Compile restricted for the mode `exec`."""
    return _compile_restricted_mode(
        source,
        filename=filename,
        mode="exec",
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy,
    )


def compile_restricted_eval(
    source,
    filename="<string>",
    flags=0,
    dont_inherit=False,
    policy=RestrictingNodeTransformer,
):
    """Compile restricted for the mode `eval`."""
    return _compile_restricted_mode(
        source,
        filename=filename,
        mode="eval",
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy,
    )


def compile_restricted_single(
    source,
    filename="<string>",
    flags=0,
    dont_inherit=False,
    policy=RestrictingNodeTransformer,
):
    """Compile restricted for the mode `single`."""
    return _compile_restricted_mode(
        source,
        filename=filename,
        mode="single",
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy,
    )


def compile_restricted_function(
    p,  # parameters
    body,
    name,
    filename="<string>",
    globalize=None,  # List of globals (e.g. ['here', 'context', ...])
    flags=0,
    dont_inherit=False,
    policy=RestrictingNodeTransformer,
):
    """Compile a restricted code object for a function.

    Documentation see:
    http://restrictedpython.readthedocs.io/en/latest/usage/index.html#RestrictedPython.compile_restricted_function
    """
    # Parse the parameters and body, then combine them.
    try:
        body_ast = ast.parse(body, "<func code>", "exec")
    except SyntaxError as v:
        error = syntax_error_template.format(
            lineno=v.lineno,
            type=v.__class__.__name__,
            msg=v.msg,
            statement=v.text.strip() if v.text else None,
        )
        return CompileResult(code=None, errors=(error,), warnings=(), used_names=())

    # The compiled code is actually executed inside a function
    # (that is called when the code is called) so reading and assigning to a
    # global variable like this`printed += 'foo'` would throw an
    # UnboundLocalError.
    # We don't want the user to need to understand this.
    if globalize:
        body_ast.body.insert(0, ast.Global(globalize))
    wrapper_ast = ast.parse(
        "def masked_function_name(%s): pass" % p, "<func wrapper>", "exec"
    )
    # In case the name you chose for your generated function is not a
    # valid python identifier we set it after the fact
    function_ast = wrapper_ast.body[0]
    assert isinstance(function_ast, ast.FunctionDef)
    function_ast.name = name

    wrapper_ast.body[0].body = body_ast.body
    wrapper_ast = ast.fix_missing_locations(wrapper_ast)

    result = _compile_restricted_mode(
        wrapper_ast,
        filename=filename,
        mode="exec",
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy,
    )

    return result


def compile_restricted(
    source,
    filename="<unknown>",
    mode="exec",
    flags=0,
    dont_inherit=False,
    policy=RestrictingNodeTransformer,
):
    """Replacement for the built-in compile() function.

    policy ... `ast.NodeTransformer` class defining the restrictions.

    """
    if mode in ["exec", "eval", "single", "function"]:
        result = _compile_restricted_mode(
            source,
            filename=filename,
            mode=mode,
            flags=flags,
            dont_inherit=dont_inherit,
            policy=policy,
        )
    else:
        raise TypeError("unknown mode %s", mode)
    for warning in result.warnings:
        warnings.warn(warning, SyntaxWarning)
    if result.errors:
        raise SyntaxError(result.errors)
    return result.code



from Aspidites._vendor.pyrsistent import (
    freeze,
    inc,
    discard,
    rex,
    ny,
    field,
    PClass,
    pmap,
)


def test_callable_command():
    m = freeze({"foo": {"bar": {"baz": 1}}})
    assert m.transform(["foo", "bar", "baz"], inc) == {"foo": {"bar": {"baz": 2}}}


def test_predicate():
    m = freeze({"foo": {"bar": {"baz": 1}, "qux": {"baz": 1}}})
    assert m.transform(["foo", lambda x: x.startswith("b"), "baz"], inc) == {
        "foo": {"bar": {"baz": 2}, "qux": {"baz": 1}}
    }


def test_broken_predicate():
    broken_predicates = [
        lambda: None,
        lambda a, b, c: None,
        lambda a, b, c, d=None: None,
        lambda *args: None,
        lambda **kwargs: None,
    ]
    for pred in broken_predicates:
        try:
            freeze({}).transform([pred], None)
            assert False
        except ValueError as e:
            assert str(e) == "callable in transform path must take 1 or 2 arguments"


def test_key_value_predicate():
    m = freeze(
        {
            "foo": 1,
            "bar": 2,
        }
    )
    assert (
        m.transform(
            [
                lambda k, v: (k, v) == ("foo", 1),
            ],
            lambda v: v * 3,
        )
        == {"foo": 3, "bar": 2}
    )


def test_remove():
    m = freeze({"foo": {"bar": {"baz": 1}}})
    assert m.transform(["foo", "bar", "baz"], discard) == {"foo": {"bar": {}}}


def test_remove_pvector():
    m = freeze({"foo": [1, 2, 3]})
    assert m.transform(["foo", 1], discard) == {"foo": [1, 3]}


def test_remove_pclass():
    class MyClass(PClass):
        a = field()
        b = field()

    m = freeze({"foo": MyClass(a=1, b=2)})
    assert m.transform(["foo", "b"], discard) == {"foo": MyClass(a=1)}


def test_predicate_no_match():
    m = freeze({"foo": {"bar": {"baz": 1}}})
    assert m.transform(["foo", lambda x: x.startswith("c"), "baz"], inc) == m


def test_rex_predicate():
    m = freeze({"foo": {"bar": {"baz": 1}, "bof": {"baz": 1}}})
    assert m.transform(["foo", rex("^bo.*"), "baz"], inc) == {
        "foo": {"bar": {"baz": 1}, "bof": {"baz": 2}}
    }


def test_rex_with_non_string_key():
    m = freeze({"foo": 1, 5: 2})
    assert m.transform([rex(".*")], 5) == {"foo": 5, 5: 2}


def test_ny_predicated_matches_any_key():
    m = freeze({"foo": 1, 5: 2})
    assert m.transform([ny], 5) == {"foo": 5, 5: 5}


def test_new_elements_created_when_missing():
    m = freeze({})
    assert m.transform(["foo", "bar", "baz"], 7) == {"foo": {"bar": {"baz": 7}}}


def test_mixed_vector_and_map():
    m = freeze({"foo": [1, 2, 3]})
    assert m.transform(["foo", 1], 5) == freeze({"foo": [1, 5, 3]})


def test_vector_predicate_callable_command():
    v = freeze([1, 2, 3, 4, 5])
    assert v.transform([lambda i: 0 < i < 4], inc) == freeze(freeze([1, 3, 4, 5, 5]))


def test_vector_insert_map_one_step_beyond_end():
    v = freeze([1, 2])
    assert v.transform([2, "foo"], 3) == freeze([1, 2, {"foo": 3}])


def test_multiple_transformations():
    v = freeze([1, 2])
    assert v.transform([2, "foo"], 3, [2, "foo"], inc) == freeze([1, 2, {"foo": 4}])


def test_no_transformation_returns_the_same_structure():
    v = freeze([{"foo": 1}, {"bar": 2}])
    assert v.transform([ny, ny], lambda x: x) is v


# def test_discard_multiple_elements_in_pvector():
#     assert freeze([0, 1, 2, 3, 4]).transform([lambda i: i % 2], discard) == freeze([0, 2, 4])


def test_transform_insert_empty_pmap():
    m = pmap().transform(["123"], pmap())
    assert m == pmap({"123": pmap()})



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected, describe_type
from ..syntax import W, add_contract, contract_expression, S, Keyword, add_keyword
from numbers import Number


class CheckType(Contract):
    def __init__(self, types, type_string=None, where=None):
        assert isinstance(types, type)
        Contract.__init__(self, where)
        self.types = types
        if type_string is None:
            self.type_string = types.__name__
        else:
            self.type_string = type_string

    def check_contract(self, context, value, silent):
        if not isinstance(value, self.types):
            error = "Expected type %r, got %s." % (
                self.types.__name__,
                describe_type(value),
            )
            raise ContractNotRespected(
                contract=self, error=error, value=value, context=context
            )

    def __str__(self):
        return self.type_string

    def __repr__(self):
        if self.types.__name__ == self.type_string:
            return "CheckType(%s)" % (self.types.__name__)
        else:
            return "CheckType(%s,%r)" % (self.types.__name__, self.type_string)

    @staticmethod
    def parse_action(types):
        def parse(s, loc, tokens):
            where = W(s, loc)
            return CheckType(types, tokens[0], where=where)  # @UnusedVariable

        return parse


add_contract(Keyword("Int").setParseAction(CheckType.parse_action(int)))
add_keyword("Int")

add_contract(Keyword("Float").setParseAction(CheckType.parse_action(float)))
add_keyword("Float")

add_contract(Keyword("Number").setParseAction(CheckType.parse_action(Number)))
add_keyword("Number")

add_contract(Keyword("bool").setParseAction(CheckType.parse_action(bool)))
add_keyword("bool")


class Type(Contract):
    def __init__(self, type_constraint, where=None):
        Contract.__init__(self, where)
        self.type_constraint = type_constraint

    def check_contract(self, context, value, silent):
        #  self.type_constraint._check_contract(context, type(value))
        self.type_constraint._check_contract(context, value.__class__, silent)

    def __str__(self):
        return "type(%s)" % self.type_constraint

    def __repr__(self):
        return "Type(%r)" % self.type_constraint

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        type_constraint = tokens["type_constraint"]
        return Type(type_constraint, where)


type_contract = (
    Keyword("type") - S("(") - contract_expression("type_constraint") - S(")")
)

add_contract(type_contract.setParseAction(Type.parse_action))
add_keyword("type")



from .func import F, curried
from .stream import Stream
from .underscore import shortcut as _
from .iters import (
    take,
    drop,
    takelast,
    droplast,
    consume,
    nth,
    first_true,
    iterate,
    padnone,
    ncycles,
    repeatfunc,
    grouper,
    group_by,
    roundrobin,
    partition,
    splitat,
    splitby,
    powerset,
    pairwise,
    iter_suppress,
    flatten,
    accumulate,
    chain,
    combinations,
    cycle,
    dropwhile,
    islice,
    repeat,
    starmap,
    takewhile,
    tee,
)
from .op import call, apply, flip, curry, zipwith, foldl, foldr, unfold

from .recur import tco, stackless

from .uniform import reduce, filterfalse, zip_longest

__version__ = "0.5.2"

__all__ = [
    "F",  # BEGIN func.py
    "curried",  # END func.py
    "take",  # BEGIN iters.py
    "drop",
    "takelast",
    "droplast",
    "consume",
    "nth",
    "first_true",
    "iterate",
    "padnone",
    "ncycles",
    "repeatfunc",
    "grouper",
    "group_by",
    "roundrobin",
    "partition",
    "splitat",
    "splitby",
    "powerset",
    "pairwise",
    "iter_suppress",
    "flatten",
    "accumulate",  # END iters.py
    "call",  # BEGIN op.py
    "apply",
    "flip",
    "curry",
    "zipwith",
    "foldl",
    "foldr",
    "unfold",  # END op.py
    "tco",
    "stackless",
    "Stream",
    "_",
    "reduce",
    "filterfalse",
    "zip_longest",
    "chain",
    "combinations",
    "cycle",
    "dropwhile",
    "islice",
    "repeat",
    "starmap",
    "takewhile",
    "tee",
]



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False

import io
import sys

from ..interface import Contract, ContractNotRespected
from ..syntax import add_contract, add_keyword, Keyword, W

# inPy2 = sys.version_info[0] == 2
# if inPy2:  # pragma: no cover
#     file_type = (file, io.IOBase)
# else:
file_type = io.IOBase


class File(Contract):
    def __init__(self, where=None):
        Contract.__init__(self, where)

    def check_contract(self, context, value, silent):

        if not isinstance(value, file_type):
            error = "Expected a file, got %r." % value.__class__.__name__
            raise ContractNotRespected(
                contract=self, error=error, value=value, context=context
            )

    def __str__(self):
        return "file"

    def __repr__(self):
        return "File()"

    @staticmethod
    def parse_action(s, loc, _):
        where = W(s, loc)
        return File(where=where)


file_contract = Keyword("file")

file_contract.setParseAction(File.parse_action)

add_contract(file_contract)
add_keyword("file")



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
import sys

from ..._compat import unicode
from ..interface import Contract, ContractNotRespected
from ..syntax import (
    add_contract,
    W,
    contract_expression,
    O,
    add_keyword,
    Keyword,
    Literal,
)


# Base class for string contracts
class StringBase(Contract):
    def __init__(self, length=None, where=None):
        Contract.__init__(self, where)
        self.length = length
        assert length is None or isinstance(length, Contract)

    def check_contract(self, context, value, silent):
        if not isinstance(value, self.TYPE):
            error = "Expected %s, got %r." % (
                self.DESCRIPTION,
                value.__class__.__name__,
            )
            raise ContractNotRespected(
                contract=self, error=error, value=value, context=context
            )

        if self.length is not None:
            self.length._check_contract(context, len(value), silent)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.length)

    def __str__(self):
        s = self.KEYWORDS[0]
        if self.length is not None:
            s += "[%s]" % self.length
        return s

    @classmethod
    def parse_action(cls, s, loc, tokens):
        where = W(s, loc)
        length = tokens.get("length", None)
        return cls(length, where=where)


if sys.version_info[0] == 3:  # Python 3

    __all__ = ["String"]

    class String(StringBase):
        KEYWORDS = ["str", "string", "unicode"]
        TYPE = str
        DESCRIPTION = "a string"


else:  # Python 2.x

    __all__ = [
        "String",
        "AnsiString",
    ]

    class String(StringBase):
        KEYWORDS = ["string"]
        TYPE = basestring
        DESCRIPTION = "an ANSI or Unicode string"

    class AnsiString(StringBase):
        KEYWORDS = ["str"]
        TYPE = str
        DESCRIPTION = "an ANSI string"


class UnicodeString(StringBase):
    KEYWORDS = ["unicode"]
    TYPE = unicode
    DESCRIPTION = "a Unicode string"


__all__ += [
    "UnicodeString",
]

for cls in StringBase.__subclasses__():
    for keyword in cls.KEYWORDS:
        mycontract = Keyword(keyword) + O(
            Literal("[") - contract_expression("length") - "]"
        )
        add_keyword(keyword)
        add_contract(mycontract.setParseAction(cls.parse_action))




# THIS FILE IS GENERATED - DO NOT EDIT #
import sysconfig
from pathlib import Path
from setuptools import setup, Extension
from Cython.Build import cythonize, BuildExecutable
from Cython.Compiler import Options
try:
    from numpy import get_include
except:
    def get_include():
        try:
            return sysconfig.get_config_vars()['CONFINCLUDEPY']
        except KeyError:
            return ''
Options.annotate=False
Options.annotate_coverage_xml=None
Options.buffer_max_dims=8
Options.cache_builtins=True
Options.cimport_from_pyx=False
Options.clear_to_none=True
Options.closure_freelist_size=8
Options.convert_range=True
Options.docstrings=True
Options.embed_pos_in_docstring=False
Options.generate_cleanup_code=False
Options.fast_fail=False
Options.warning_errors=False
Options.error_on_unknown_names=True
Options.error_on_uninitialized=True
Options.gcc_branch_hints=True
Options.lookup_module_cpdef=False
Options.embed=None
ext = cythonize('Aspidites/woma/library.pyx', include_path=[get_include()], quiet=True)
setup(name='Aspidites.woma.library',ext_modules=ext)
if Options.embed: BuildExecutable.build('Aspidites/woma/library.pyx')




from . import good, fail


class BaseClass:
    pass


class SubClass(BaseClass):
    pass


good("isinstance(BaseClass)", BaseClass())
good("isinstance(BaseClass)", SubClass())
good("isinstance(SubClass)", SubClass())
fail("isinstance(SubClass)", BaseClass())


class BaseClass2(object):
    pass


class SubClass2(BaseClass2):
    pass


good("isinstance(BaseClass2)", BaseClass2())
good("isinstance(BaseClass2)", SubClass2())
good("isinstance(SubClass2)", SubClass2())
fail("isinstance(SubClass2)", BaseClass2())

# Let's try with 3, old-style


class BaseClass3:
    pass


class MidClass3(BaseClass3):
    pass


class SubClass3(MidClass3):
    pass


good("isinstance(BaseClass3)", BaseClass3())
good("isinstance(BaseClass3)", SubClass3())
good("isinstance(SubClass3)", SubClass3())
fail("isinstance(SubClass3)", BaseClass3())


class BaseClass4(object):
    pass


class MidClass4(BaseClass4):
    pass


class SubClass4(MidClass4):
    pass


good("isinstance(BaseClass4)", BaseClass4())
good("isinstance(BaseClass4)", SubClass4())
good("isinstance(SubClass4)", SubClass4())
fail("isinstance(SubClass4)", BaseClass4())
